#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import io
from typing import Optional

from ..serialization import MessageField, SerializationContext, SerializationError
from .schema_registry_client import (
    AsyncSchemaRegistryClient,
    ConfigCompatibilityLevel,
    Metadata,
    MetadataProperties,
    MetadataTags,
    RegisteredSchema,
    Rule,
    RuleKind,
    RuleMode,
    RuleParams,
    RuleSet,
    Schema,
    SchemaReference,
    SchemaRegistryClient,
    SchemaRegistryError,
    ServerConfig,
)

_KEY_SCHEMA_ID = "__key_schema_id"
_VALUE_SCHEMA_ID = "__value_schema_id"

_MAGIC_BYTE = 0
_MAGIC_BYTE_V0 = _MAGIC_BYTE
_MAGIC_BYTE_V1 = 1

__all__ = [
    "ConfigCompatibilityLevel",
    "Metadata",
    "MetadataProperties",
    "MetadataTags",
    "RegisteredSchema",
    "Rule",
    "RuleKind",
    "RuleMode",
    "RuleParams",
    "RuleSet",
    "Schema",
    "SchemaRegistryClient",
    "AsyncSchemaRegistryClient",
    "SchemaRegistryError",
    "SchemaReference",
    "ServerConfig",
    "topic_subject_name_strategy",
    "topic_record_subject_name_strategy",
    "record_subject_name_strategy",
    "header_schema_id_serializer",
    "prefix_schema_id_serializer",
    "dual_schema_id_deserializer",
    "prefix_schema_id_deserializer",
]


def topic_subject_name_strategy(ctx: Optional[SerializationContext], record_name: Optional[str]) -> Optional[str]:
    """
    Constructs a subject name in the form of {topic}-key|value.

    Args:
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation. **Required** - will raise ValueError if None.

        record_name (Optional[str]): Record name.

    Raises:
        ValueError: If ctx is None.

    """
    if ctx is None:
        raise ValueError(
            "SerializationContext is required for topic_subject_name_strategy. "
            "Either provide a SerializationContext or use record_subject_name_strategy."
        )
    return ctx.topic + "-" + ctx.field


def topic_record_subject_name_strategy(
    ctx: Optional[SerializationContext], record_name: Optional[str]
) -> Optional[str]:
    """
    Constructs a subject name in the form of {topic}-{record_name}.

    Args:
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation. **Required** - will raise ValueError if None.

        record_name (Optional[str]): Record name.

    Raises:
        ValueError: If ctx is None.

    """
    if ctx is None:
        raise ValueError(
            "SerializationContext is required for topic_record_subject_name_strategy. "
            "Either provide a SerializationContext or use record_subject_name_strategy."
        )
    return ctx.topic + "-" + record_name if record_name is not None else None


def record_subject_name_strategy(ctx: Optional[SerializationContext], record_name: Optional[str]) -> Optional[str]:
    """
    Constructs a subject name in the form of {record_name}.

    Args:
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation. **Not used** by this strategy.

        record_name (Optional[str]): Record name.

    Note:
        This strategy does not require SerializationContext and can be used
        when ctx is None.

    """
    return record_name if record_name is not None else None


def reference_subject_name_strategy(ctx: Optional[SerializationContext], schema_ref: SchemaReference) -> Optional[str]:
    """
    Constructs a subject reference name in the form of {reference name}.

    Args:
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation. **Not used** by this strategy.

        schema_ref (SchemaReference): SchemaReference instance.

    Note:
        This strategy does not require SerializationContext and can be used
        when ctx is None.

    """
    return schema_ref.name if schema_ref is not None else None


def header_schema_id_serializer(payload: bytes, ctx: Optional[SerializationContext], schema_id) -> bytes:
    """
    Serializes the schema guid into the header.

    Args:
        payload (bytes): The payload to serialize.
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation.
        schema_id (SchemaId): The schema ID to serialize.

    Returns:
        bytes: The payload
    """
    if ctx is None:
        raise SerializationError("SerializationContext is required for header_schema_id_serializer")

    headers = ctx.headers
    if headers is None:
        raise SerializationError("Missing headers")
    header_key = _KEY_SCHEMA_ID if ctx.field == MessageField.KEY else _VALUE_SCHEMA_ID
    header_value = schema_id.guid_to_bytes()
    if isinstance(headers, list):
        headers.append((header_key, header_value))
    elif isinstance(headers, dict):
        headers[header_key] = header_value
    else:
        raise SerializationError("Invalid headers type")
    return payload


def prefix_schema_id_serializer(payload: bytes, ctx, schema_id) -> bytes:
    """
    Serializes the schema id into the payload prefix.

    Args:
        payload (bytes): The payload to serialize.
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation.
        schema_id (SchemaId): The schema ID to serialize.

    Returns:
        bytes: The payload prefixed with the schema id
    """
    return schema_id.id_to_bytes() + payload


def dual_schema_id_deserializer(payload: bytes, ctx: Optional[SerializationContext], schema_id) -> io.BytesIO:
    """
    Deserializes the schema id by first checking the header, then the payload prefix.

    Args:
        payload (bytes): The payload to serialize.
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation.
        schema_id (SchemaId): The schema ID to serialize.

    Returns:
        bytes: The payload
    """
    # Look for schema ID in headers
    header_value = None

    if ctx is not None:
        headers = ctx.headers
        if headers is not None:
            header_key = _KEY_SCHEMA_ID if ctx.field == MessageField.KEY else _VALUE_SCHEMA_ID
            if isinstance(headers, list):
                # look for header_key in headers
                for header in headers:
                    if header[0] == header_key:
                        header_value = header[1]
                        break
            elif isinstance(headers, dict):
                header_value = headers.get(header_key, None)

    # Parse schema ID from determined source and return appropriate payload
    if header_value is not None:
        schema_id.from_bytes(io.BytesIO(header_value))  # type: ignore[arg-type]
        return io.BytesIO(payload)  # Return full payload when schema ID is in header
    else:
        return schema_id.from_bytes(io.BytesIO(payload))  # Parse from payload, return remainder


def prefix_schema_id_deserializer(payload: bytes, ctx, schema_id) -> io.BytesIO:
    """
    Deserializes the schema id from the payload prefix.

    Args:
        payload (bytes): The payload to serialize.
        ctx (SerializationContext): Metadata pertaining to the serialization
            operation.
        schema_id (SchemaId): The schema ID to serialize.

    Returns:
        bytes: The payload
    """
    return schema_id.from_bytes(io.BytesIO(payload))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
from collections import defaultdict
from enum import Enum
from threading import Lock
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

__all__ = [
    'VALID_AUTH_PROVIDERS',
    'is_success',
    'is_retriable',
    'full_jitter',
    '_SchemaCache',
    'RuleKind',
    'RuleMode',
    'RuleParams',
    'Rule',
    'RuleSet',
    'MetadataTags',
    'MetadataProperties',
    'Metadata',
    'SchemaReference',
    'ConfigCompatibilityLevel',
    'ServerConfig',
    'Schema',
    'RegisteredSchema',
]

VALID_AUTH_PROVIDERS = ['URL', 'USER_INFO']


def is_success(status_code: int) -> bool:
    return 200 <= status_code <= 299


def is_retriable(status_code: int) -> bool:
    return status_code in (408, 429, 500, 502, 503, 504)


def full_jitter(base_delay_ms: int, max_delay_ms: int, retries_attempted: int) -> float:
    no_jitter_delay = base_delay_ms * (2.0**retries_attempted)
    return random.random() * min(no_jitter_delay, max_delay_ms)


class _SchemaCache(object):
    """
    Thread-safe cache for use with the Schema Registry Client.

    This cache may be used to retrieve schema ids, schemas or to check
    known subject membership.
    """

    def __init__(self):
        self.lock = Lock()
        self.schema_id_index = defaultdict(dict)
        self.schema_guid_index = {}
        self.schema_index = defaultdict(dict)
        self.rs_id_index = defaultdict(dict)
        self.rs_version_index = defaultdict(dict)
        self.rs_schema_index = defaultdict(dict)

    def set_schema(self, subject: Optional[str], schema_id: Optional[int], guid: Optional[str], schema: 'Schema'):
        """
        Add a Schema identified by schema_id to the cache.

        Args:
            subject (str): The subject this schema is associated with

            schema_id (int): Schema's id

            guid (str): Schema's guid

            schema (Schema): Schema instance
        """

        with self.lock:
            if schema_id is not None:
                self.schema_id_index[subject][schema_id] = (guid, schema)
                self.schema_index[subject][schema] = schema_id
            if guid is not None:
                self.schema_guid_index[guid] = schema

    def set_registered_schema(self, schema: 'Schema', registered_schema: 'RegisteredSchema'):
        """
        Add a RegisteredSchema to the cache.

        Args:
            schema (Schema): Schema instance
            registered_schema (RegisteredSchema): RegisteredSchema instance
        """

        subject = registered_schema.subject
        schema_id = registered_schema.schema_id
        guid = registered_schema.guid
        version = registered_schema.version
        with self.lock:
            if schema_id is not None:
                self.schema_id_index[subject][schema_id] = (guid, schema)
                self.schema_index[subject][schema] = schema_id
                self.rs_id_index[subject][schema_id] = registered_schema
            if guid is not None:
                self.schema_guid_index[guid] = schema
            self.rs_version_index[subject][version] = registered_schema
            self.rs_schema_index[subject][schema] = registered_schema

    def get_schema_by_id(self, subject: Optional[str], schema_id: int) -> Optional[Tuple[str, 'Schema']]:
        """
        Get the schema instance associated with schema id from the cache.

        Args:
            subject (str): The subject this schema is associated with

            schema_id (int): Id used to identify a schema

        Returns:
            Tuple[str, Schema]: The guid and schema if known; else None
        """

        with self.lock:
            return self.schema_id_index.get(subject, {}).get(schema_id, None)

    def get_schema_by_guid(self, guid: str) -> Optional['Schema']:
        """
        Get the schema instance associated with guid from the cache.

        Args:
            guid (str): Guid used to identify a schema

        Returns:
            Schema: The schema if known; else None
        """

        with self.lock:
            return self.schema_guid_index.get(guid, None)

    def get_id_by_schema(self, subject: str, schema: 'Schema') -> Optional[int]:
        """
        Get the schema id associated with schema instance from the cache.

        Args:
            subject (str): The subject this schema is associated with

            schema (Schema): The schema

        Returns:
            int: The schema id if known; else None
        """

        with self.lock:
            return self.schema_index.get(subject, {}).get(schema, None)

    def get_registered_by_subject_schema(self, subject: str, schema: 'Schema') -> Optional['RegisteredSchema']:
        """
        Get the schema associated with this schema registered under subject.

        Args:
            subject (str): The subject this schema is associated with

            schema (Schema): The schema associated with this schema

        Returns:
            RegisteredSchema: The registered schema if known; else None
        """

        with self.lock:
            return self.rs_schema_index.get(subject, {}).get(schema, None)

    def get_registered_by_subject_id(self, subject: str, schema_id: int) -> Optional['RegisteredSchema']:
        """
        Get the schema associated with this id registered under subject.

        Args:
            subject (str): The subject this schema is associated with

            schema_id (int): The schema id associated with this schema

        Returns:
            RegisteredSchema: The registered schema if known; else None
        """

        with self.lock:
            return self.rs_id_index.get(subject, {}).get(schema_id, None)

    def get_registered_by_subject_version(self, subject: str, version: int) -> Optional['RegisteredSchema']:
        """
        Get the schema associated with this version registered under subject.

        Args:
            subject (str): The subject this schema is associated with

            version (int): The version associated with this schema

        Returns:
            RegisteredSchema: The registered schema if known; else None
        """

        with self.lock:
            return self.rs_version_index.get(subject, {}).get(version, None)

    def remove_by_subject(self, subject: str):
        """
        Remove schemas with the given subject.

        Args:
            subject (str): The subject
        """

        with self.lock:
            if subject in self.schema_id_index:
                del self.schema_id_index[subject]
            if subject in self.schema_index:
                del self.schema_index[subject]
            if subject in self.rs_id_index:
                del self.rs_id_index[subject]
            if subject in self.rs_version_index:
                del self.rs_version_index[subject]
            if subject in self.rs_schema_index:
                del self.rs_schema_index[subject]

    def remove_by_subject_version(self, subject: str, version: int):
        """
        Remove schemas with the given subject.

        Args:
            subject (str): The subject

            version (int) The version
        """

        with self.lock:
            if subject in self.rs_id_index:
                for schema_id, registered_schema in list(self.rs_id_index[subject].items()):
                    if registered_schema.version == version:
                        del self.rs_id_index[subject][schema_id]

            if subject in self.rs_schema_index:
                for schema, registered_schema in list(self.rs_schema_index[subject].items()):
                    if registered_schema.version == version:
                        del self.rs_schema_index[subject][schema]
            rs = None
            if subject in self.rs_version_index:
                if version in self.rs_version_index[subject]:
                    rs = self.rs_version_index[subject][version]
                    del self.rs_version_index[subject][version]
            if rs is not None:
                if subject in self.schema_id_index:
                    if rs.schema_id in self.schema_id_index[subject]:
                        del self.schema_id_index[subject][rs.schema_id]
                if subject in self.schema_index:
                    if rs.schema in self.schema_index[subject]:
                        del self.schema_index[subject][rs.schema]

    def clear(self):
        """
        Clear the cache.
        """

        with self.lock:
            self.schema_id_index.clear()
            self.schema_guid_index.clear()
            self.schema_index.clear()
            self.rs_id_index.clear()
            self.rs_version_index.clear()
            self.rs_schema_index.clear()


T = TypeVar("T")


class RuleKind(str, Enum):
    CONDITION = "CONDITION"
    TRANSFORM = "TRANSFORM"

    def __str__(self) -> str:
        return str(self.value)


class RulePhase(str, Enum):
    MIGRATION = "MIGRATION"
    DOMAIN = "DOMAIN"
    ENCODING = "ENCODING"

    def __str__(self) -> str:
        return str(self.value)


class RuleMode(str, Enum):
    UPGRADE = "UPGRADE"
    DOWNGRADE = "DOWNGRADE"
    UPDOWN = "UPDOWN"
    READ = "READ"
    WRITE = "WRITE"
    WRITEREAD = "WRITEREAD"

    def __str__(self) -> str:
        return str(self.value)


@_attrs_define
class RuleParams:
    params: Dict[str, str] = _attrs_field(factory=dict, hash=False)

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        field_dict.update(self.params)

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        rule_params = cls(params=d)  # type: ignore[call-arg]

        return rule_params

    def __hash__(self):
        return hash(frozenset(self.params.items()))


@_attrs_define(frozen=True)
class Rule:
    name: Optional[str]
    doc: Optional[str]
    kind: Optional[RuleKind]
    mode: Optional[RuleMode]
    type: Optional[str]
    tags: Optional[List[str]] = _attrs_field(hash=False)
    params: Optional[RuleParams]
    expr: Optional[str]
    on_success: Optional[str]
    on_failure: Optional[str]
    disabled: Optional[bool]

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        doc = self.doc

        kind_str: Optional[str] = None
        if self.kind is not None:
            kind_str = self.kind.value

        mode_str: Optional[str] = None
        if self.mode is not None:
            mode_str = self.mode.value

        rule_type = self.type

        tags = self.tags

        _params: Optional[Dict[str, Any]] = None
        if self.params is not None:
            _params = self.params.to_dict()

        expr = self.expr

        on_success = self.on_success

        on_failure = self.on_failure

        disabled = self.disabled

        field_dict: Dict[str, Any] = {}
        field_dict.update({})
        if name is not None:
            field_dict["name"] = name
        if doc is not None:
            field_dict["doc"] = doc
        if kind_str is not None:
            field_dict["kind"] = kind_str
        if mode_str is not None:
            field_dict["mode"] = mode_str
        if type is not None:
            field_dict["type"] = rule_type
        if tags is not None:
            field_dict["tags"] = tags
        if _params is not None:
            field_dict["params"] = _params
        if expr is not None:
            field_dict["expr"] = expr
        if on_success is not None:
            field_dict["onSuccess"] = on_success
        if on_failure is not None:
            field_dict["onFailure"] = on_failure
        if disabled is not None:
            field_dict["disabled"] = disabled

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name", None)

        doc = d.pop("doc", None)

        _kind = d.pop("kind", None)
        kind: Optional[RuleKind] = None
        if _kind is not None:
            kind = RuleKind(_kind)

        _mode = d.pop("mode", None)
        mode: Optional[RuleMode] = None
        if _mode is not None:
            mode = RuleMode(_mode)

        rule_type = d.pop("type", None)

        tags = cast(List[str], d.pop("tags", None))

        _params: Optional[Dict[str, Any]] = d.pop("params", None)
        params: Optional[RuleParams] = None
        if _params is not None:
            params = RuleParams.from_dict(_params)

        expr = d.pop("expr", None)

        on_success = d.pop("onSuccess", None)

        on_failure = d.pop("onFailure", None)

        disabled = d.pop("disabled", None)

        rule = cls(  # type: ignore[call-arg]
            name=name,
            doc=doc,
            kind=kind,
            mode=mode,
            type=rule_type,
            tags=tags,
            params=params,
            expr=expr,
            on_success=on_success,
            on_failure=on_failure,
            disabled=disabled,
        )

        return rule


@_attrs_define
class RuleSet:
    migration_rules: Optional[List["Rule"]] = _attrs_field(hash=False)
    domain_rules: Optional[List["Rule"]] = _attrs_field(hash=False)
    encoding_rules: Optional[List["Rule"]] = _attrs_field(hash=False, default=None)

    def to_dict(self) -> Dict[str, Any]:
        _migration_rules: Optional[List[Dict[str, Any]]] = None
        if self.migration_rules is not None:
            _migration_rules = []
            for migration_rules_item_data in self.migration_rules:
                migration_rules_item = migration_rules_item_data.to_dict()
                _migration_rules.append(migration_rules_item)

        _domain_rules: Optional[List[Dict[str, Any]]] = None
        if self.domain_rules is not None:
            _domain_rules = []
            for domain_rules_item_data in self.domain_rules:
                domain_rules_item = domain_rules_item_data.to_dict()
                _domain_rules.append(domain_rules_item)

        _encoding_rules: Optional[List[Dict[str, Any]]] = None
        if self.encoding_rules is not None:
            _encoding_rules = []
            for encoding_rules_item_data in self.encoding_rules:
                encoding_rules_item = encoding_rules_item_data.to_dict()
                _encoding_rules.append(encoding_rules_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update({})
        if _migration_rules is not None:
            field_dict["migrationRules"] = _migration_rules
        if _domain_rules is not None:
            field_dict["domainRules"] = _domain_rules
        if _encoding_rules is not None:
            field_dict["encodingRules"] = _encoding_rules

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        migration_rules = []
        _migration_rules = d.pop("migrationRules", None)
        for migration_rules_item_data in _migration_rules or []:
            migration_rules_item = Rule.from_dict(migration_rules_item_data)
            migration_rules.append(migration_rules_item)

        domain_rules = []
        _domain_rules = d.pop("domainRules", None)
        for domain_rules_item_data in _domain_rules or []:
            domain_rules_item = Rule.from_dict(domain_rules_item_data)
            domain_rules.append(domain_rules_item)

        encoding_rules = []
        _encoding_rules = d.pop("encodingRules", None)
        for encoding_rules_item_data in _encoding_rules or []:
            encoding_rules_item = Rule.from_dict(encoding_rules_item_data)
            encoding_rules.append(encoding_rules_item)

        rule_set = cls(  # type: ignore[call-arg]
            migration_rules=migration_rules,
            domain_rules=domain_rules,
            encoding_rules=encoding_rules,
        )

        return rule_set

    def __hash__(self):
        return hash(frozenset((self.migration_rules or []) + (self.domain_rules or []) + (self.encoding_rules or [])))


@_attrs_define
class MetadataTags:
    tags: Dict[str, List[str]] = _attrs_field(factory=dict, hash=False)

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        for prop_name, prop in self.tags.items():
            field_dict[prop_name] = prop

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        tags = {}
        for prop_name, prop_dict in d.items():
            tag = cast(List[str], prop_dict)

            tags[prop_name] = tag

        metadata_tags = cls(tags=tags)  # type: ignore[call-arg]

        return metadata_tags

    def __hash__(self):
        return hash(frozenset(self.tags.items()))


@_attrs_define
class MetadataProperties:
    properties: Dict[str, str] = _attrs_field(factory=dict, hash=False)

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        field_dict.update(self.properties)

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        metadata_properties = cls(properties=d)  # type: ignore[call-arg]

        return metadata_properties

    def __hash__(self):
        return hash(frozenset(self.properties.items()))


@_attrs_define(frozen=True)
class Metadata:
    tags: Optional[MetadataTags]
    properties: Optional[MetadataProperties]
    sensitive: Optional[List[str]] = _attrs_field(hash=False)

    def to_dict(self) -> Dict[str, Any]:
        _tags: Optional[Dict[str, Any]] = None
        if self.tags is not None:
            _tags = self.tags.to_dict()

        _properties: Optional[Dict[str, Any]] = None
        if self.properties is not None:
            _properties = self.properties.to_dict()

        sensitive: Optional[List[str]] = None
        if self.sensitive is not None:
            sensitive = []
            for sensitive_item in self.sensitive:
                sensitive.append(sensitive_item)

        field_dict: Dict[str, Any] = {}
        if _tags is not None:
            field_dict["tags"] = _tags
        if _properties is not None:
            field_dict["properties"] = _properties
        if sensitive is not None:
            field_dict["sensitive"] = sensitive

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        _tags: Optional[Dict[str, Any]] = d.pop("tags", None)
        tags: Optional[MetadataTags] = None
        if _tags is not None:
            tags = MetadataTags.from_dict(_tags)

        _properties: Optional[Dict[str, Any]] = d.pop("properties", None)
        properties: Optional[MetadataProperties] = None
        if _properties is not None:
            properties = MetadataProperties.from_dict(_properties)

        sensitive = []
        _sensitive = d.pop("sensitive", None)
        for sensitive_item in _sensitive or []:
            sensitive.append(sensitive_item)

        metadata = cls(  # type: ignore[call-arg]
            tags=tags,
            properties=properties,
            sensitive=sensitive,
        )

        return metadata


@_attrs_define(frozen=True)
class SchemaVersion:
    subject: Optional[str]
    version: Optional[int]

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        return cls(subject=src_dict.get('subject'), version=src_dict.get('version'))  # type: ignore[call-arg]


@_attrs_define(frozen=True)
class SchemaReference:
    name: Optional[str]
    subject: Optional[str]
    version: Optional[int]

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        subject = self.subject

        version = self.version

        field_dict: Dict[str, Any] = {}
        if name is not None:
            field_dict["name"] = name
        if subject is not None:
            field_dict["subject"] = subject
        if version is not None:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        return cls(  # type: ignore[call-arg]
            name=src_dict.get('name'),
            subject=src_dict.get('subject'),
            version=src_dict.get('version'),
        )


class ConfigCompatibilityLevel(str, Enum):
    BACKWARD = "BACKWARD"
    BACKWARD_TRANSITIVE = "BACKWARD_TRANSITIVE"
    FORWARD = "FORWARD"
    FORWARD_TRANSITIVE = "FORWARD_TRANSITIVE"
    FULL = "FULL"
    FULL_TRANSITIVE = "FULL_TRANSITIVE"
    NONE = "NONE"

    def __str__(self) -> str:
        return str(self.value)


@_attrs_define
class ServerConfig:
    compatibility: Optional[ConfigCompatibilityLevel] = None
    compatibility_level: Optional[ConfigCompatibilityLevel] = None
    compatibility_group: Optional[str] = None
    default_metadata: Optional[Metadata] = None
    override_metadata: Optional[Metadata] = None
    default_rule_set: Optional[RuleSet] = None
    override_rule_set: Optional[RuleSet] = None

    def to_dict(self) -> Dict[str, Any]:
        _compatibility: Optional[str] = None
        if self.compatibility is not None:
            _compatibility = self.compatibility.value

        _compatibility_level: Optional[str] = None
        if self.compatibility_level is not None:
            _compatibility_level = self.compatibility_level.value

        compatibility_group = self.compatibility_group

        _default_metadata: Optional[Dict[str, Any]]
        if isinstance(self.default_metadata, Metadata):
            _default_metadata = self.default_metadata.to_dict()
        else:
            _default_metadata = self.default_metadata

        _override_metadata: Optional[Dict[str, Any]]
        if isinstance(self.override_metadata, Metadata):
            _override_metadata = self.override_metadata.to_dict()
        else:
            _override_metadata = self.override_metadata

        _default_rule_set: Optional[Dict[str, Any]]
        if isinstance(self.default_rule_set, RuleSet):
            _default_rule_set = self.default_rule_set.to_dict()
        else:
            _default_rule_set = self.default_rule_set

        _override_rule_set: Optional[Dict[str, Any]]
        if isinstance(self.override_rule_set, RuleSet):
            _override_rule_set = self.override_rule_set.to_dict()
        else:
            _override_rule_set = self.override_rule_set

        field_dict: Dict[str, Any] = {}
        if _compatibility is not None:
            field_dict["compatibility"] = _compatibility
        if _compatibility_level is not None:
            field_dict["compatibilityLevel"] = _compatibility_level
        if compatibility_group is not None:
            field_dict["compatibilityGroup"] = compatibility_group
        if _default_metadata is not None:
            field_dict["defaultMetadata"] = _default_metadata
        if _override_metadata is not None:
            field_dict["overrideMetadata"] = _override_metadata
        if _default_rule_set is not None:
            field_dict["defaultRuleSet"] = _default_rule_set
        if _override_rule_set is not None:
            field_dict["overrideRuleSet"] = _override_rule_set

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        _compatibility = d.pop("compatibility", None)
        compatibility: Optional[ConfigCompatibilityLevel]
        if _compatibility is None:
            compatibility = None
        else:
            compatibility = ConfigCompatibilityLevel(_compatibility)

        _compatibility_level = d.pop("compatibilityLevel", None)
        compatibility_level: Optional[ConfigCompatibilityLevel]
        if _compatibility_level is None:
            compatibility_level = None
        else:
            compatibility_level = ConfigCompatibilityLevel(_compatibility_level)

        compatibility_group = d.pop("compatibilityGroup", None)

        def _parse_default_metadata(data: object) -> Optional[Metadata]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return Metadata.from_dict(data)

        default_metadata = _parse_default_metadata(d.pop("defaultMetadata", None))

        def _parse_override_metadata(data: object) -> Optional[Metadata]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return Metadata.from_dict(data)

        override_metadata = _parse_override_metadata(d.pop("overrideMetadata", None))

        def _parse_default_rule_set(data: object) -> Optional[RuleSet]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return RuleSet.from_dict(data)

        default_rule_set = _parse_default_rule_set(d.pop("defaultRuleSet", None))

        def _parse_override_rule_set(data: object) -> Optional[RuleSet]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return RuleSet.from_dict(data)

        override_rule_set = _parse_override_rule_set(d.pop("overrideRuleSet", None))

        config = cls(  # type: ignore[call-arg]
            compatibility=compatibility,
            compatibility_level=compatibility_level,
            compatibility_group=compatibility_group,
            default_metadata=default_metadata,
            override_metadata=override_metadata,
            default_rule_set=default_rule_set,
            override_rule_set=override_rule_set,
        )

        return config


@_attrs_define(frozen=True, cache_hash=True)
class Schema:
    """
    An unregistered schema.
    """

    schema_str: Optional[str]
    schema_type: Optional[str] = "AVRO"
    references: Optional[List[SchemaReference]] = _attrs_field(factory=list, hash=False)
    metadata: Optional[Metadata] = None
    rule_set: Optional[RuleSet] = None

    def to_dict(self) -> Dict[str, Any]:
        schema = self.schema_str
        schema_type = self.schema_type

        _references: Optional[List[Dict[str, Any]]] = []
        if self.references is not None:
            for references_item_data in self.references:
                references_item = references_item_data.to_dict()
                _references.append(references_item)  # type: ignore[union-attr]

        _metadata: Optional[Dict[str, Any]] = None
        if isinstance(self.metadata, Metadata):
            _metadata = self.metadata.to_dict()

        _rule_set: Optional[Dict[str, Any]] = None
        if isinstance(self.rule_set, RuleSet):
            _rule_set = self.rule_set.to_dict()

        field_dict: Dict[str, Any] = {}
        if schema is not None:
            field_dict["schema"] = schema
        if schema_type is not None:
            field_dict["schemaType"] = schema_type
        if _references is not None:
            field_dict["references"] = _references
        if _metadata is not None:
            field_dict["metadata"] = _metadata
        if _rule_set is not None:
            field_dict["ruleSet"] = _rule_set

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        schema = d.pop("schema", None)

        schema_type = d.pop("schemaType", "AVRO")

        references = []
        _references = d.pop("references", None)
        for references_item_data in _references or []:
            references_item = SchemaReference.from_dict(references_item_data)

            references.append(references_item)

        def _parse_metadata(data: object) -> Optional[Metadata]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return Metadata.from_dict(data)

        metadata = _parse_metadata(d.pop("metadata", None))

        def _parse_rule_set(data: object) -> Optional[RuleSet]:
            if data is None:
                return data
            if not isinstance(data, dict):
                raise TypeError()
            return RuleSet.from_dict(data)

        rule_set = _parse_rule_set(d.pop("ruleSet", None))

        schema = cls(  # type: ignore[call-arg]
            schema_str=schema,
            schema_type=schema_type,
            references=references,
            metadata=metadata,
            rule_set=rule_set,
        )

        return schema


@_attrs_define(frozen=True, cache_hash=True)
class RegisteredSchema:
    """
    An registered schema.
    """

    subject: Optional[str]
    version: Optional[int]
    schema_id: Optional[int]
    guid: Optional[str]
    schema: Schema

    def to_dict(self) -> Dict[str, Any]:
        schema = self.schema

        schema_id = self.schema_id

        guid = self.guid

        subject = self.subject

        version = self.version

        field_dict: Dict[str, Any] = {}
        if schema is not None:
            field_dict = schema.to_dict()
        if schema_id is not None:
            field_dict["id"] = schema_id
        if guid is not None:
            field_dict["guid"] = guid
        if subject is not None:
            field_dict["subject"] = subject
        if version is not None:
            field_dict["version"] = version

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        schema = Schema.from_dict(d)

        schema_id = d.pop("id", None)

        guid = d.pop("guid", None)

        subject = d.pop("subject", None)

        version = d.pop("version", None)

        schema = cls(  # type: ignore[call-arg,assignment]
            schema_id=schema_id,
            guid=guid,
            schema=schema,
            subject=subject,
            version=version,
        )

        return schema  # type: ignore[return-value]



# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import functools
import logging
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar

T = TypeVar('T')


class AsyncLogger:

    def __init__(self, loop: asyncio.AbstractEventLoop, logger: logging.Logger) -> None:
        self.loop = loop
        self.logger = logger

    def log(self, *args: Any, **kwargs: Any) -> None:
        self.loop.call_soon_threadsafe(lambda: self.logger.log(*args, **kwargs))


def wrap_callback(
    loop: asyncio.AbstractEventLoop,
    callback: Callable[..., Any],
    edit_args: Optional[Callable[[Tuple[Any, ...]], Tuple[Any, ...]]] = None,
    edit_kwargs: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
) -> Callable[..., Any]:
    def ret(*args: Any, **kwargs: Any) -> Any:
        if edit_args:
            args = edit_args(args)
        if edit_kwargs:
            kwargs = edit_kwargs(kwargs)
        f = asyncio.run_coroutine_threadsafe(callback(*args, **kwargs), loop)
        return f.result()

    return ret


def wrap_conf_callback(loop: asyncio.AbstractEventLoop, conf: Dict[str, Any], name: str) -> None:
    if name in conf:
        cb = conf[name]
        conf[name] = wrap_callback(loop, cb)


def wrap_conf_logger(loop: asyncio.AbstractEventLoop, conf: Dict[str, Any]) -> None:
    if 'logger' in conf:
        conf['logger'] = AsyncLogger(loop, conf['logger'])


async def async_call(
    executor: concurrent.futures.Executor, blocking_task: Callable[..., T], *args: Any, **kwargs: Any
) -> T:
    """Helper function for blocking operations that need ThreadPool execution

    Args:
        executor: ThreadPoolExecutor to use for blocking operations
        blocking_task: The blocking function to execute
        *args, **kwargs: Arguments to pass to the blocking function

    Returns:
        Result of the blocking function execution
    """
    return (
        await asyncio.gather(
            asyncio.get_running_loop().run_in_executor(executor, functools.partial(blocking_task, *args, **kwargs))
        )
    )[0]


def wrap_common_callbacks(loop: asyncio.AbstractEventLoop, conf: Dict[str, Any]) -> None:
    wrap_conf_callback(loop, conf, 'error_cb')
    wrap_conf_callback(loop, conf, 'throttle_cb')
    wrap_conf_callback(loop, conf, 'stats_cb')
    wrap_conf_callback(loop, conf, 'oauth_cb')
    wrap_conf_logger(loop, conf)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2026 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc
from urllib.parse import urlparse, urlunparse

__all__ = [
    '_AbstractOAuthBearerFieldProviderBuilder',
    '_AbstractOAuthBearerOIDCFieldProviderBuilder',
    '_StaticOAuthBearerFieldProviderBuilder',
    '_AbstractCustomOAuthBearerFieldProviderBuilder',
    '_AbstractOAuthBearerOIDCAzureIMDSFieldProviderBuilder',
    '_BearerFieldProvider',
    '_AsyncBearerFieldProvider',
]


class _BearerFieldProvider(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_bearer_fields(self) -> dict:
        raise NotImplementedError


class _AsyncBearerFieldProvider(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def get_bearer_fields(self) -> dict:
        raise NotImplementedError


class _AbstractOAuthBearerFieldProviderBuilder(metaclass=abc.ABCMeta):
    """Abstract base class for OAuthBearer client builders"""

    required_properties = ['bearer.auth.logical.cluster', 'bearer.auth.identity.pool.id']

    def __init__(self, conf: dict):
        self.conf: dict = conf
        self.logical_cluster: str = ""
        self.identity_pool: str = ""

    def _validate(self):
        missing_properties = [
            prop for prop in _AbstractOAuthBearerFieldProviderBuilder.required_properties if prop not in self.conf
        ]
        if missing_properties:
            raise ValueError(
                "Missing required bearer configuration properties: {}".format(", ".join(missing_properties))
            )

        self.logical_cluster = self.conf.pop('bearer.auth.logical.cluster')
        if not isinstance(self.logical_cluster, str):
            raise TypeError("logical cluster must be a str, not " + str(type(self.logical_cluster)))

        self.identity_pool = self.conf.pop('bearer.auth.identity.pool.id')
        if not isinstance(self.identity_pool, str):
            raise TypeError("identity pool id must be a str, not " + str(type(self.identity_pool)))

    @abc.abstractmethod
    def build(self, max_retries: int, retries_wait_ms: int, retries_max_wait_ms: int) -> _BearerFieldProvider:
        pass


class _AbstractOAuthBearerOIDCFieldProviderBuilder(_AbstractOAuthBearerFieldProviderBuilder):
    required_properties = [
        'bearer.auth.client.id',
        'bearer.auth.client.secret',
        'bearer.auth.scope',
        'bearer.auth.issuer.endpoint.url',
    ]

    def __init__(self, conf: dict):
        super().__init__(conf)
        self.client_id: str = ""
        self.client_secret: str = ""
        self.scope: str = ""
        self.token_endpoint: str = ""

    def _validate(self):
        super()._validate()

        missing_properties = [
            prop for prop in _AbstractOAuthBearerOIDCFieldProviderBuilder.required_properties if prop not in self.conf
        ]
        if missing_properties:
            raise ValueError(
                "Missing required OAuth configuration properties: {}".format(", ".join(missing_properties))
            )

        self.client_id = self.conf.pop('bearer.auth.client.id')
        if not isinstance(self.client_id, str):
            raise TypeError("bearer.auth.client.id must be a str, not " + str(type(self.client_id)))

        self.client_secret = self.conf.pop('bearer.auth.client.secret')
        if not isinstance(self.client_secret, str):
            raise TypeError("bearer.auth.client.secret must be a str, not " + str(type(self.client_secret)))

        self.scope = self.conf.pop('bearer.auth.scope')
        if not isinstance(self.scope, str):
            raise TypeError("bearer.auth.scope must be a str, not " + str(type(self.scope)))

        self.token_endpoint = self.conf.pop('bearer.auth.issuer.endpoint.url')
        if not isinstance(self.token_endpoint, str):
            raise TypeError("bearer.auth.issuer.endpoint.url must be a str, not " + str(type(self.token_endpoint)))


class _AbstractOAuthBearerOIDCAzureIMDSFieldProviderBuilder(_AbstractOAuthBearerFieldProviderBuilder):

    def __init__(self, conf: dict):
        super().__init__(conf)
        self.token_endpoint: str = 'http://169.254.169.254/metadata/identity/oauth2/token'

    def _validate(self):
        super()._validate()

        token_endpoint_override = 'bearer.auth.issuer.endpoint.url' in self.conf
        self.token_endpoint = self.conf.pop('bearer.auth.issuer.endpoint.url', self.token_endpoint)
        if not isinstance(self.token_endpoint, str):
            raise TypeError("bearer.auth.issuer.endpoint.url must be a str, not " + str(type(self.token_endpoint)))

        try:
            parsed_token_endpoint = urlparse(self.token_endpoint)
        except Exception as ex:
            raise ValueError(f'Failed to parse token endpoint URL: {ex}')

        token_query = self.conf.pop('bearer.auth.issuer.endpoint.query', None)
        if token_query:
            if not isinstance(token_query, str):
                raise TypeError("bearer.auth.issuer.endpoint.query must be a str, not " + str(type(token_query)))

            parsed_token_endpoint = parsed_token_endpoint._replace(query=token_query, fragment=None)
            self.token_endpoint = urlunparse(parsed_token_endpoint)
        elif not token_endpoint_override:
            raise ValueError(
                "bearer.auth.issuer.endpoint.query must be provided "
                "when bearer.auth.issuer.endpoint.url isn't overridden"
            )


class _StaticFieldProvider(_BearerFieldProvider):
    def __init__(self, token: str, logical_cluster: str, identity_pool: str):
        self.token: str = token
        self.logical_cluster: str = logical_cluster
        self.identity_pool: str = identity_pool

    def get_bearer_fields(self) -> dict:
        return {
            'bearer.auth.token': self.token,
            'bearer.auth.logical.cluster': self.logical_cluster,
            'bearer.auth.identity.pool.id': self.identity_pool,
        }


class _StaticOAuthBearerFieldProviderBuilder(_AbstractOAuthBearerFieldProviderBuilder):

    def __init__(self, conf: dict):
        super().__init__(conf)
        self.static_token: str = ""

    def _validate(self):
        super()._validate()

        if 'bearer.auth.token' not in self.conf:
            raise ValueError("Missing bearer.auth.token")
        self.static_token = self.conf.pop('bearer.auth.token')
        if not isinstance(self.static_token, str):
            raise TypeError("bearer.auth.token must be a str, not " + str(type(self.static_token)))

    def build(self, max_retries: int, retries_wait_ms: int, retries_max_wait_ms: int):
        self._validate()
        return _StaticFieldProvider(self.static_token, self.logical_cluster, self.identity_pool)


class _AbstractCustomOAuthBearerFieldProviderBuilder:
    required_properties = ['bearer.auth.custom.provider.function', 'bearer.auth.custom.provider.config']

    def __init__(self, conf: dict):
        self.conf = conf
        self.custom_function = None
        self.custom_config = None

    def _validate(self):
        missing_properties = [
            prop for prop in _AbstractCustomOAuthBearerFieldProviderBuilder.required_properties if prop not in self.conf
        ]
        if missing_properties:
            raise ValueError(
                "Missing required custom OAuth configuration properties: {}".format(", ".join(missing_properties))
            )

        self.custom_function = self.conf.pop('bearer.auth.custom.provider.function')
        if not callable(self.custom_function):
            raise TypeError(
                "bearer.auth.custom.provider.function must be a callable, not " + str(type(self.custom_function))
            )

        self.custom_config = self.conf.pop('bearer.auth.custom.provider.config')
        if not isinstance(self.custom_config, dict):
            raise TypeError("bearer.auth.custom.provider.config must be a dict, not " + str(type(self.custom_config)))



# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
import weakref
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    # Import only for type checking to avoid circular dependency
    from ._kafka_batch_executor import ProducerBatchExecutor
    from ._producer_batch_processor import ProducerBatchManager

logger = logging.getLogger(__name__)


class BufferTimeoutManager:
    """Manages buffer timeout and activity tracking for message batching

    This class is responsible for:
    - Monitoring buffer inactivity and triggering automatic flushes
    - Tracking buffer activity timestamps
    - Managing background timeout monitoring tasks
    - Coordinating between batch processor and executor for timeout flushes
    """

    def __init__(
        self, batch_processor: "ProducerBatchManager", kafka_executor: "ProducerBatchExecutor", timeout: float
    ) -> None:
        """Initialize the buffer timeout manager

        Args:
            batch_processor: ProducerBatchManager instance for creating batches
            kafka_executor: ProducerBatchExecutor instance for executing batches
            timeout: Timeout in seconds for buffer inactivity (0 disables timeout)
        """
        self._batch_processor = batch_processor
        self._kafka_executor = kafka_executor
        self._timeout = timeout
        self._last_activity: float = time.time()
        self._timeout_task: Optional[asyncio.Task[None]] = None
        self._running: bool = False

    def start_timeout_monitoring(self) -> None:
        """Start the background task that monitors buffer inactivity

        Creates an async task that runs in the background and periodically checks
        if messages have been sitting in the buffer for too long without being
        flushed.

        Key design decisions:
        1. **Weak Reference**: Uses weakref.ref(self) to prevent circular refs
        2. **Self-Canceling**: The task stops itself if manager is GC'd
        3. **Adaptive Check Interval**: Uses timeout to determine check frequency
        """
        if not self._timeout or self._timeout <= 0:
            return  # Timeout disabled

        self._running = True
        self._timeout_task = asyncio.create_task(self._monitor_timeout())

    def stop_timeout_monitoring(self) -> None:
        """Stop and cleanup the buffer timeout monitoring task"""
        self._running = False
        if self._timeout_task and not self._timeout_task.done():
            self._timeout_task.cancel()
            self._timeout_task = None

    def mark_activity(self) -> None:
        """Update the timestamp of the last buffer activity

        This method should be called whenever:
        1. Messages are added to the buffer (in produce())
        2. Buffer is manually flushed
        3. Buffer is purged/cleared
        """
        self._last_activity = time.time()

    async def _monitor_timeout(self) -> None:
        """Monitor buffer timeout in background task

        This method runs continuously in the background, checking for buffer
        inactivity and triggering flushes when the timeout threshold is exceeded.
        """
        # Use weak reference to avoid circular reference and allow garbage collection
        manager_ref = weakref.ref(self)

        while True:
            # Check interval should be proportional to buffer timeout for efficiency
            manager = manager_ref()
            if manager is None or not manager._running:
                break

            # Calculate adaptive check interval: timeout/2 with bounds
            # Examples: 0.1s→0.1s, 1s→0.5s, 5s→1.0s, 30s→1.0s
            check_interval = max(0.1, min(1.0, manager._timeout / 2))
            await asyncio.sleep(check_interval)

            # Re-check manager after sleep
            manager = manager_ref()
            if manager is None or not manager._running:
                break

            # Check if buffer has been inactive for too long
            time_since_activity = time.time() - manager._last_activity
            if time_since_activity >= manager._timeout:

                try:
                    # Flush the buffer due to timeout
                    await manager._flush_buffer_due_to_timeout()
                    # Update activity since we just flushed
                    manager.mark_activity()
                except Exception:
                    logger.error("Error flushing buffer due to timeout", exc_info=True)
                    # Re-raise all exceptions - don't swallow any errors
                    raise

    async def _flush_buffer_due_to_timeout(self) -> None:
        """Flush buffer due to timeout by coordinating batch processor/executor

        This method handles the complete timeout flush workflow:
        1. Create batches from the batch processor
        2. Execute batches from the batch processor
        3. Flush librdkafka queue to ensure messages are delivered
        """
        # Create batches from current buffer and send to librdkafka queue
        await self._batch_processor.flush_buffer()

        # Flush librdkafka queue to ensure messages are delivered to broker
        # 0 timeout means non-blocking flush
        await self._kafka_executor.flush_librdkafka_queue(0)



# Copyright 2023 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, Optional

from .. import cimpl


class OffsetSpec(ABC):
    """
    Used in `AdminClient.list_offsets` to specify the desired offsets
    of the partition being queried.
    """

    _values: Dict[int, 'OffsetSpec'] = {}
    _max_timestamp: Optional['MaxTimestampSpec'] = None
    _earliest: Optional['EarliestSpec'] = None
    _latest: Optional['LatestSpec'] = None

    @property
    @abstractmethod
    def _value(self) -> int:
        pass

    @classmethod
    def _fill_values(cls) -> None:
        cls._max_timestamp = MaxTimestampSpec()
        cls._earliest = EarliestSpec()
        cls._latest = LatestSpec()
        cls._values.update(
            {
                cimpl.OFFSET_SPEC_MAX_TIMESTAMP: cls._max_timestamp,
                cimpl.OFFSET_SPEC_EARLIEST: cls._earliest,
                cimpl.OFFSET_SPEC_LATEST: cls._latest,
            }
        )

    @classmethod
    def earliest(cls):
        return cls._earliest

    @classmethod
    def latest(cls):
        return cls._latest

    @classmethod
    def max_timestamp(cls):
        return cls._max_timestamp

    @classmethod
    def for_timestamp(cls, timestamp: int):
        return TimestampSpec(timestamp)

    def __new__(cls, index: int):
        # Trying to instantiate returns one of the subclasses.
        # Subclasses can be instantiated but aren't accessible externally.
        if index < 0:
            return cls._values[index]
        else:
            return cls.for_timestamp(index)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, OffsetSpec):
            return NotImplemented
        return self._value < other._value


class TimestampSpec(OffsetSpec):
    """
    Used in a `AdminClient.list_offsets` call to retrieve the earliest offset
    whose timestamp is greater than or equal to the given timestamp in the
    corresponding partition.

    Parameters
    ----------
    timestamp: int
        timestamp in milliseconds.
    """

    @property
    def _value(self) -> int:
        return self.timestamp

    def __new__(cls, _: int):
        return object.__new__(cls)

    def __init__(self, timestamp: int) -> None:
        self.timestamp = timestamp


class MaxTimestampSpec(OffsetSpec):
    """
    Used in a `AdminClient.list_offsets` call to retrieve the offset with the
    largest timestamp, that could not correspond to the latest one as timestamps
    can be specified client-side.
    """

    def __new__(cls):
        return object.__new__(cls)

    @property
    def _value(self) -> int:
        return cimpl.OFFSET_SPEC_MAX_TIMESTAMP


class LatestSpec(OffsetSpec):
    """
    Used in a `AdminClient.list_offsets` call to retrieve the queried partition latest offset.
    """

    def __new__(cls):
        return object.__new__(cls)

    @property
    def _value(self) -> int:
        return cimpl.OFFSET_SPEC_LATEST


class EarliestSpec(OffsetSpec):
    """
    Used in a `AdminClient.list_offsets` call to retrieve the queried partition earliest offset.
    """

    def __new__(cls):
        return object.__new__(cls)

    @property
    def _value(self) -> int:
        return cimpl.OFFSET_SPEC_EARLIEST


OffsetSpec._fill_values()


class ListOffsetsResultInfo:
    """
    ListOffsetsResultInfo
    Result of a `AdminClient.list_offsets` call associated to a partition.

    Parameters
    ----------
    offset: int
        The offset returned by the list_offsets call.
    timestamp: int
        The timestamp in milliseconds corresponding to the offset.
        Not available (-1) when querying for the earliest or the latest offsets.
    leader_epoch: int
        The leader epoch corresponding to the offset (optional).
    """

    def __init__(self, offset: int, timestamp: int, leader_epoch: int) -> None:
        self.offset = offset
        self.timestamp = timestamp
        self.leader_epoch: Optional[int] = leader_epoch
        if leader_epoch < 0:
            self.leader_epoch = None



# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

from .._model import Node
from .._util import ConversionUtil
from ._acl import AclOperation


class DescribeClusterResult:
    """
    Represents cluster description information used in describe cluster operation.
    Used by :meth:`AdminClient.describe_cluster`.

    Parameters
    ----------
    controller : Node
        The current controller in the cluster.
    nodes : list(Node)
        Information about each node in the cluster.
    cluster_id : str
        The current cluster id in the cluster.
    authorized_operations: list(AclOperation)
        AclOperations allowed for the cluster.
    """

    def __init__(
        self,
        controller: Node,
        nodes: List[Node],
        cluster_id: Optional[str] = None,
        authorized_operations: Optional[List[Union[str, int, AclOperation]]] = None,
    ) -> None:
        self.cluster_id = cluster_id
        self.controller = controller
        self.nodes = nodes
        self.authorized_operations = None
        if authorized_operations:
            self.authorized_operations = []
            for op in authorized_operations:
                self.authorized_operations.append(ConversionUtil.convert_to_enum(op, AclOperation))



#!/usr/bin/env python
#
# Copyright 2017 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class ClientError(Exception):
    """Error thrown by Schema Registry clients"""

    def __init__(self, message, http_code=None):
        self.message = message
        self.http_code = http_code
        super(ClientError, self).__init__(self.__str__())

    def __repr__(self):
        return "ClientError(error={error})".format(error=self.message)

    def __str__(self):
        return self.message



# Copyright 2025 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Tuple

from confluent_kafka import KafkaException as _KafkaException

from ._message_batch import MessageBatch, create_message_batch

if TYPE_CHECKING:
    # Import only for type checking to avoid circular dependency
    from ._kafka_batch_executor import ProducerBatchExecutor

logger = logging.getLogger(__name__)


class ProducerBatchManager:
    """Handles batching and processing of Kafka messages for AIOProducer

    This class encapsulates all the logic for:
    - Grouping messages by topic and partition
    - Managing message buffers and futures
    - Creating simple future-resolving callbacks
    - Executing batch operations via librdkafka
    """

    def __init__(self, kafka_executor: "ProducerBatchExecutor") -> None:
        """Initialize the batch processor

        Args:
            kafka_executor: KafkaBatchExecutor instance for Kafka operations
        """
        self._kafka_executor = kafka_executor
        self._message_buffer: List[Dict[str, Any]] = []
        self._buffer_futures: List[asyncio.Future[Any]] = []

    def add_message(self, msg_data: Dict[str, Any], future: asyncio.Future[Any]) -> None:
        """Add a message to the batch buffer

        Args:
            msg_data: Dictionary containing message data
            future: asyncio.Future to resolve when message is delivered
        """
        self._message_buffer.append(msg_data)
        self._buffer_futures.append(future)

    def get_buffer_size(self) -> int:
        """Get the current number of messages in the buffer"""
        return len(self._message_buffer)

    def is_buffer_empty(self) -> bool:
        """Check if the buffer is empty"""
        return len(self._message_buffer) == 0

    def clear_buffer(self) -> None:
        """Clear the entire buffer"""
        self._message_buffer.clear()
        self._buffer_futures.clear()

    def cancel_pending_futures(self) -> None:
        """Cancel all pending futures in the buffer"""
        for future in self._buffer_futures:
            if not future.done():
                future.cancel()

    def create_batches(self, target_topic: Optional[str] = None) -> List[MessageBatch]:
        """Create MessageBatch objects from the current buffer

        Args:
            target_topic: Optional topic to create batches for (None for all)

        Returns:
            List[MessageBatch]: List of immutable MessageBatch objects
        """
        if self.is_buffer_empty():
            return []

        # Group by topic and partition for optimal batching
        topic_partition_groups = self._group_messages_by_topic_and_partition()
        batches = []

        for (topic, partition), group_data in topic_partition_groups.items():
            if target_topic is None or topic == target_topic:
                # Prepare batch messages
                batch_messages = self._prepare_batch_messages(group_data["messages"])

                # Assign simple future-resolving callbacks to messages
                self._assign_future_callbacks(batch_messages, group_data["futures"])

                # Create immutable MessageBatch object with partition info
                batch = create_message_batch(
                    topic=topic,
                    messages=batch_messages,
                    futures=group_data["futures"],
                    callbacks=None,  # No user callbacks anymore
                    partition=partition,  # Add partition info to batch
                )
                batches.append(batch)

        return batches

    def _clear_topic_from_buffer(self, target_topic: str) -> None:
        """Remove messages for a specific topic from the buffer

        Args:
            target_topic: Topic to remove from buffer
        """
        messages_to_keep = []
        futures_to_keep = []

        for i, msg_data in enumerate(self._message_buffer):
            if msg_data["topic"] != target_topic:
                messages_to_keep.append(msg_data)
                futures_to_keep.append(self._buffer_futures[i])

        self._message_buffer = messages_to_keep
        self._buffer_futures = futures_to_keep

    async def flush_buffer(self, target_topic: Optional[str] = None) -> None:
        """Flush the current message buffer using produce_batch

        Args:
            target_topic: Optional topic to flush (None for all topics)

        Returns:
            None
        """
        if self.is_buffer_empty():
            return

        # Create batches for processing
        batches = self.create_batches(target_topic)

        # Clear the buffer immediately to prevent race conditions
        if target_topic is None:
            # Clear entire buffer since we're processing all messages
            self.clear_buffer()
        else:
            # Clear only messages for the target topic that we're processing
            self._clear_topic_from_buffer(target_topic)

        try:
            # Execute batches with cleanup
            await self._execute_batches(batches, target_topic)
        except Exception:
            # Add batches back to buffer on failure
            try:
                self._add_batches_back_to_buffer(batches)
            except Exception:
                logger.error(f"Error adding batches back to buffer on failure. messages might be lost: {batches}")
                raise
            raise

    async def _execute_batches(self, batches: List[MessageBatch], target_topic: Optional[str] = None) -> None:
        """Execute batches and handle cleanup after successful execution

        Args:
            batches: List of batches to execute
            target_topic: Optional topic for selective buffer clearing

        Returns:
            None

        Raises:
            Exception: If any batch execution fails
        """
        # Execute each batch
        for batch in batches:
            try:
                # Execute batch using the Kafka executor
                await self._kafka_executor.execute_batch(batch.topic, batch.messages, batch.partition)

            except Exception as e:
                # Handle batch failure by failing all unresolved futures for this batch
                self._handle_batch_failure(e, batch.futures)
                # Re-raise the exception so caller knows the batch operation failed
                raise

    def _add_batches_back_to_buffer(self, batches: List[MessageBatch]) -> None:
        """Add batches back to the buffer when execution fails

        Args:
            batches: List of MessageBatch objects to add back to buffer
        """
        for batch in batches:
            # Add each message and its future back to the buffer
            for i, message in enumerate(batch.messages):
                # Reconstruct the original message data from the batch
                msg_data = {
                    'topic': batch.topic,
                    'value': message.get('value'),
                    'key': message.get('key'),
                }

                # Add optional fields if present
                if 'partition' in message:
                    msg_data['partition'] = message['partition']
                if 'timestamp' in message:
                    msg_data['timestamp'] = message['timestamp']
                if 'headers' in message:
                    msg_data['headers'] = message['headers']

                # Add the message and its future back to the buffer
                self._message_buffer.append(msg_data)
                self._buffer_futures.append(batch.futures[i])

    def _group_messages_by_topic_and_partition(self) -> Dict[Tuple[str, int], Dict[str, List[Any]]]:
        """Group buffered messages by topic and partition for optimal batching

        This function efficiently organizes the mixed-topic message buffer into
        topic+partition-specific groups, enabling proper partition control while
        maintaining batch efficiency.

        Algorithm:
        - Single O(n) pass through message buffer
        - Groups related data (messages, futures) by (topic, partition) tuple
        - Maintains index relationships between buffer arrays
        - Uses partition from message data, defaults to RD_KAFKA_PARTITION_UA
          (-1) if not specified

        Returns:
            dict: Topic+partition groups with structure:
                {
                    ('topic_name', partition): {
                        'messages': [msg_data1, ...],  # Message dicts
                        'futures': [future1, ...],     # asyncio.Future objects
                    }
                }
        """
        topic_partition_groups: Dict[Tuple[str, int], Dict[str, Any]] = {}

        # Iterate through buffer once - O(n) complexity
        for i, msg_data in enumerate(self._message_buffer):
            topic = msg_data["topic"]
            # Get partition from message data, default to RD_KAFKA_PARTITION_UA (-1) if not specified
            partition = msg_data.get("partition", -1)  # -1 = RD_KAFKA_PARTITION_UA

            # Create composite key for grouping
            group_key = (topic, partition)

            # Create new topic+partition group if this is first message for this combination
            if group_key not in topic_partition_groups:
                topic_partition_groups[group_key] = {
                    "messages": [],  # Message data for produce_batch
                    "futures": [],  # Futures to resolve on delivery
                }

            # Add message and related data to appropriate topic+partition group
            # Note: All arrays stay synchronized by index
            topic_partition_groups[group_key]["messages"].append(msg_data)
            topic_partition_groups[group_key]["futures"].append(self._buffer_futures[i])

        return topic_partition_groups

    def _prepare_batch_messages(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Prepare messages for produce_batch by removing internal fields

        Args:
            messages: List of message dictionaries

        Returns:
            List of cleaned message dictionaries ready for produce_batch
        """
        batch_messages = []
        for msg_data in messages:
            # Create a shallow copy and remove fields not needed by produce_batch
            batch_msg = copy.copy(msg_data)
            batch_msg.pop("topic", None)  # Remove topic since it's passed separately
            # Note: We keep 'partition' in individual messages for reference,
            # but the batch partition will be used by produce_batch
            batch_messages.append(batch_msg)

        return batch_messages

    def _assign_future_callbacks(
        self, batch_messages: List[Dict[str, Any]], futures: Sequence[asyncio.Future[Any]]
    ) -> None:
        """Assign simple future-resolving callbacks to each message in batch

        Args:
            batch_messages: List of message dictionaries for produce_batch
            futures: List of asyncio.Future objects to resolve
        """
        for i, batch_msg in enumerate(batch_messages):
            future = futures[i]

            def create_simple_callback(fut: asyncio.Future[Any]) -> Callable[[Any, Any], None]:
                """Create a simple callback that only resolves the future"""

                def simple_callback(err: Any, msg: Any) -> None:
                    if err:
                        if not fut.done():
                            fut.set_exception(_KafkaException(err))
                    else:
                        if not fut.done():
                            fut.set_result(msg)

                return simple_callback

            # Assign the simple callback to this message
            batch_msg["callback"] = create_simple_callback(future)

    def _handle_batch_failure(self, exception: Exception, batch_futures: Sequence[asyncio.Future[Any]]) -> None:
        """Handle batch operation failure by failing all unresolved futures

        When a batch operation fails before any individual callbacks are invoked,
        we need to fail all futures for this batch since none of the per-message
        callbacks will be called by librdkafka.

        Args:
            exception: The exception that caused the batch to fail
            batch_futures: List of futures for this batch
        """
        # Fail all futures since no individual callbacks will be invoked
        for future in batch_futures:
            # Only set exception if future isn't already done
            if not future.done():
                future.set_exception(exception)



# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A client for Google Cloud KMS."""

import tink
from azure.keyvault.keys.crypto import CryptographyClient, EncryptionAlgorithm
from tink import aead


class AzureKmsAead(aead.Aead):
    """Implements the Aead interface for Azure KMS."""

    def __init__(self, client: CryptographyClient, algorithm: EncryptionAlgorithm) -> None:
        if not client:
            raise tink.TinkError('client cannot be null.')
        self.client = client
        self.algorithm = algorithm

    def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
        try:
            response = self.client.encrypt(self.algorithm, plaintext)
            return response.ciphertext
        except ValueError as e:
            raise tink.TinkError(e)

    def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
        try:
            response = self.client.decrypt(self.algorithm, ciphertext)
            return response.plaintext
        except ValueError as e:
            raise tink.TinkError(e)



# Copyright 2024 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A client for Google Cloud KMS."""

from typing import Optional
from urllib.parse import urlparse

import hvac
import tink
from tink import aead
from tink.integration.hcvault import new_aead

VAULT_KEYURI_PREFIX = 'hcvault://'


class HcVaultKmsClient(tink.KmsClient):
    """Basic HashiCorp Vault client for AEAD."""

    def __init__(
        self,
        key_uri: str,
        token: Optional[str],
        ns: Optional[str] = None,
        role_id: Optional[str] = None,
        secret_id: Optional[str] = None,
    ) -> None:
        """Creates a new HcVaultKmsClient that is bound to the key specified in 'key_uri'.

        Uses the specified credentials when communicating with the KMS.

        Args:
          key_uri: The URI of the key the client should be bound to.
          token: The Vault token.
          ns: The Vault namespace.

        Raises:
          TinkError: If the key uri is not valid.
        """

        if key_uri.startswith(VAULT_KEYURI_PREFIX):
            self._key_uri = key_uri
        else:
            raise tink.TinkError('Invalid key_uri.')

        parsed = urlparse(key_uri[len(VAULT_KEYURI_PREFIX) :])
        vault_url = parsed.scheme + '://' + parsed.netloc
        self._client = hvac.Client(url=vault_url, token=token, namespace=ns, verify=False)
        if role_id and secret_id and self._client is not None:
            self._client.auth.approle.login(role_id=role_id, secret_id=secret_id)

    def does_support(self, key_uri: str) -> bool:
        """Returns true iff this client supports KMS key specified in 'key_uri'.

        Args:
          key_uri: URI of the key to be checked.

        Returns:
          A boolean value which is true if the key is supported and false otherwise.
        """
        if not self._key_uri:
            return key_uri.startswith(VAULT_KEYURI_PREFIX)
        return key_uri == self._key_uri

    def get_aead(self, key_uri: str) -> aead.Aead:
        """Returns an Aead-primitive backed by KMS key specified by 'key_uri'.

        Args:
          key_uri: URI of the key which should be used.

        Returns:
          An Aead object.
        """
        if self._key_uri and self._key_uri != key_uri:
            raise tink.TinkError('This client is bound to %s and cannot use key %s' % (self._key_uri, key_uri))
        if not key_uri.startswith(VAULT_KEYURI_PREFIX):
            raise tink.TinkError('Invalid key_uri.')
        key_id = key_uri[len(VAULT_KEYURI_PREFIX) :]
        parsed = urlparse(key_id)
        return new_aead(parsed.path, self._client)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: confluent/meta.proto
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x14\x63onfluent/meta.proto\x12\tconfluent\x1a google/protobuf/descriptor.proto\"}\n\x04Meta\x12\x0b\n\x03\x64oc\x18\x01 \x01(\t\x12+\n\x06params\x18\x02 \x03(\x0b\x32\x1b.confluent.Meta.ParamsEntry\x12\x0c\n\x04tags\x18\x03 \x03(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01:A\n\tfile_meta\x12\x1c.google.protobuf.FileOptions\x18\xc0\x08 \x01(\x0b\x32\x0f.confluent.Meta:G\n\x0cmessage_meta\x12\x1f.google.protobuf.MessageOptions\x18\xc0\x08 \x01(\x0b\x32\x0f.confluent.Meta:C\n\nfield_meta\x12\x1d.google.protobuf.FieldOptions\x18\xc0\x08 \x01(\x0b\x32\x0f.confluent.Meta:A\n\tenum_meta\x12\x1c.google.protobuf.EnumOptions\x18\xc0\x08 \x01(\x0b\x32\x0f.confluent.Meta:L\n\x0f\x65num_value_meta\x12!.google.protobuf.EnumValueOptions\x18\xc0\x08 \x01(\x0b\x32\x0f.confluent.MetaB\x0eZ\x0c../confluentb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'confluent.meta_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(file_meta)
    google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(message_meta)
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(field_meta)
    google_dot_protobuf_dot_descriptor__pb2.EnumOptions.RegisterExtension(enum_meta)
    google_dot_protobuf_dot_descriptor__pb2.EnumValueOptions.RegisterExtension(enum_value_meta)

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'Z\014../confluent'
    _META_PARAMSENTRY._options = None
    _META_PARAMSENTRY._serialized_options = b'8\001'
    _META._serialized_start = 69
    _META._serialized_end = 194
    _META_PARAMSENTRY._serialized_start = 149
    _META_PARAMSENTRY._serialized_end = 194
# @@protoc_insertion_point(module_scope)



# Copyright 2022 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Kafka admin client: create, view, alter, and delete topics and resources.

Note: Many imports in this file are marked with "# noqa: F401" because they are
intentionally unused within this module but are exported as part of the public API.
These imports allow users to access constants and classes directly from the admin
module (e.g., "from confluent_kafka.admin import CONFIG_SOURCE_DEFAULT_CONFIG").
"""

import concurrent.futures
import warnings
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from confluent_kafka import ConsumerGroupState as _ConsumerGroupState
from confluent_kafka import ConsumerGroupTopicPartitions as _ConsumerGroupTopicPartitions
from confluent_kafka import IsolationLevel as _IsolationLevel

from .._model import ConsumerGroupType as _ConsumerGroupType
from .._model import ElectionType as _ElectionType
from .._model import TopicCollection as _TopicCollection
from ..cimpl import KafkaException  # noqa: F401
from ..cimpl import (  # noqa: F401
    CONFIG_SOURCE_DEFAULT_CONFIG,
    CONFIG_SOURCE_DYNAMIC_BROKER_CONFIG,
    CONFIG_SOURCE_DYNAMIC_DEFAULT_BROKER_CONFIG,
    CONFIG_SOURCE_DYNAMIC_TOPIC_CONFIG,
    CONFIG_SOURCE_GROUP_CONFIG,
    CONFIG_SOURCE_STATIC_BROKER_CONFIG,
    CONFIG_SOURCE_UNKNOWN_CONFIG,
    OFFSET_INVALID,
    RESOURCE_ANY,
    RESOURCE_BROKER,
    RESOURCE_GROUP,
    RESOURCE_TOPIC,
    RESOURCE_TRANSACTIONAL_ID,
    RESOURCE_UNKNOWN,
    KafkaError,
    NewPartitions,
    NewTopic,
)
from ..cimpl import TopicPartition as _TopicPartition
from ..cimpl import (  # noqa: F401
    _AdminClientImpl,
)
from ._acl import AclOperation  # noqa: F401
from ._acl import AclBinding, AclBindingFilter, AclPermissionType  # noqa: F401
from ._cluster import DescribeClusterResult  # noqa: F401

# Unused imports are keeped to be accessible using this public module
from ._config import ConfigSource  # noqa: F401
from ._config import AlterConfigOpType, ConfigEntry, ConfigResource  # noqa: F401
from ._group import ConsumerGroupListing  # noqa: F401
from ._group import (  # noqa: F401
    ConsumerGroupDescription,
    ListConsumerGroupsResult,
    MemberAssignment,
    MemberDescription,
)
from ._listoffsets import ListOffsetsResultInfo  # noqa: F401
from ._listoffsets import OffsetSpec  # noqa: F401
from ._metadata import BrokerMetadata  # noqa: F401
from ._metadata import ClusterMetadata, GroupMember, GroupMetadata, PartitionMetadata, TopicMetadata  # noqa: F401
from ._records import DeletedRecords  # noqa: F401
from ._resource import ResourcePatternType  # noqa: F401
from ._resource import ResourceType  # noqa: F401
from ._scram import UserScramCredentialAlteration  # noqa: F401
from ._scram import (  # noqa: F401
    ScramCredentialInfo,
    ScramMechanism,
    UserScramCredentialDeletion,
    UserScramCredentialsDescription,
    UserScramCredentialUpsertion,
)
from ._topic import TopicDescription  # noqa: F401

try:
    string_type = basestring  # type: ignore[name-defined]
except NameError:
    string_type = str


class AdminClient(_AdminClientImpl):
    """
    AdminClient provides admin operations for Kafka brokers, topics, groups,
    and other resource types supported by the broker.

    The Admin API methods are asynchronous and return a dict of
    concurrent.futures.Future objects keyed by the entity.
    The entity is a topic name for create_topics(), delete_topics(), create_partitions(),
    and a ConfigResource for alter_configs() and describe_configs().

    All the futures for a single API call will currently finish/fail at
    the same time (backed by the same protocol request), but this might
    change in future versions of the client.

    See examples/adminapi.py for example usage.

    For more information see the `Java Admin API documentation
    <https://docs.confluent.io/current/clients/javadocs/org/apache/kafka/clients/admin/package-frame.html>`_.

    Requires broker version v0.11.0.0 or later.
    """

    def __init__(self, conf: Dict[str, Union[str, int, float, bool]], **kwargs: Any) -> None:
        """
        Create a new AdminClient using the provided configuration dictionary.

        The AdminClient is a standard Kafka protocol client, supporting
        the standard librdkafka configuration properties as specified at
        https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md

        :param dict conf: Configuration properties. At a minimum ``bootstrap.servers`` **should** be set\n"
        :param Logger logger: Optional Logger instance to use as a custom log messages handler.
        """
        super(AdminClient, self).__init__(conf, **kwargs)

    @staticmethod
    def _make_topics_result(f: concurrent.futures.Future, futmap: Dict[str, concurrent.futures.Future]) -> None:
        """
        Map per-topic results to per-topic futures in futmap.
        The result value of each (successful) future is None.
        """
        try:
            result = f.result()
            for topic, error in result.items():
                fut = futmap.get(topic, None)
                if fut is None:
                    raise RuntimeError("Topic {} not found in future-map: {}".format(topic, futmap))

                if error is not None:
                    # Topic-level exception
                    fut.set_exception(KafkaException(error))
                else:
                    # Topic-level success
                    fut.set_result(None)
        except Exception as e:
            # Request-level exception, raise the same for all topics
            for topic, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_resource_result(
        f: concurrent.futures.Future, futmap: Dict[ConfigResource, concurrent.futures.Future]
    ) -> None:
        """
        Map per-resource results to per-resource futures in futmap.
        The result value of each (successful) future is a ConfigResource.
        """
        try:
            result = f.result()
            for resource, configs in result.items():
                fut = futmap.get(resource, None)
                if fut is None:
                    raise RuntimeError("Resource {} not found in future-map: {}".format(resource, futmap))
                if resource.error is not None:
                    # Resource-level exception
                    fut.set_exception(KafkaException(resource.error))
                else:
                    # Resource-level success
                    # configs will be a dict for describe_configs()
                    # and None for alter_configs()
                    fut.set_result(configs)
        except Exception as e:
            # Request-level exception, raise the same for all resources
            for resource, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_list_consumer_groups_result(f: concurrent.futures.Future, futmap: Any) -> None:
        pass

    @staticmethod
    def _make_consumer_groups_result(
        f: concurrent.futures.Future, futmap: Dict[str, concurrent.futures.Future]
    ) -> None:
        """
        Map per-group results to per-group futures in futmap.
        """
        try:

            results = f.result()
            futmap_values = list(futmap.values())
            len_results = len(results)
            len_futures = len(futmap_values)
            if len_results != len_futures:
                raise RuntimeError(
                    "Results length {} is different from future-map length {}".format(len_results, len_futures)
                )
            for i, result in enumerate(results):
                fut = futmap_values[i]
                if isinstance(result, KafkaError):
                    fut.set_exception(KafkaException(result))
                else:
                    fut.set_result(result)
        except Exception as e:
            # Request-level exception, raise the same for all groups
            for _, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_consumer_group_offsets_result(
        f: concurrent.futures.Future, futmap: Dict[str, concurrent.futures.Future]
    ) -> None:
        """
        Map per-group results to per-group futures in futmap.
        The result value of each (successful) future is ConsumerGroupTopicPartitions.
        """
        try:

            results = f.result()
            futmap_values = list(futmap.values())
            len_results = len(results)
            len_futures = len(futmap_values)
            if len_results != len_futures:
                raise RuntimeError(
                    "Results length {} is different from future-map length {}".format(len_results, len_futures)
                )
            for i, result in enumerate(results):
                fut = futmap_values[i]
                if isinstance(result, KafkaError):
                    fut.set_exception(KafkaException(result))
                else:
                    fut.set_result(result)
        except Exception as e:
            # Request-level exception, raise the same for all groups
            for _, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_acls_result(f: concurrent.futures.Future, futmap: Dict[Any, concurrent.futures.Future]) -> None:
        """
        Map create ACL binding results to corresponding futures in futmap.
        For create_acls the result value of each (successful) future is None.
        For delete_acls the result value of each (successful) future is the list of deleted AclBindings.
        """
        try:
            results = f.result()
            futmap_values = list(futmap.values())
            len_results = len(results)
            len_futures = len(futmap_values)
            if len_results != len_futures:
                raise RuntimeError(
                    "Results length {} is different from future-map length {}".format(len_results, len_futures)
                )
            for i, result in enumerate(results):
                fut = futmap_values[i]
                if isinstance(result, KafkaError):
                    fut.set_exception(KafkaException(result))
                else:
                    fut.set_result(result)
        except Exception as e:
            # Request-level exception, raise the same for all the AclBindings or AclBindingFilters
            for resource, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_futmap_result_from_list(
        f: concurrent.futures.Future, futmap: Dict[Any, concurrent.futures.Future]
    ) -> None:
        try:

            results = f.result()
            futmap_values = list(futmap.values())
            len_results = len(results)
            len_futures = len(futmap_values)
            if len_results != len_futures:
                raise RuntimeError(
                    "Results length {} is different from future-map length {}".format(len_results, len_futures)
                )
            for i, result in enumerate(results):
                fut = futmap_values[i]
                if isinstance(result, KafkaError):
                    fut.set_exception(KafkaException(result))
                else:
                    fut.set_result(result)
        except Exception as e:
            # Request-level exception, raise the same for all topics
            for _, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _make_futmap_result(f: concurrent.futures.Future, futmap: Dict[str, concurrent.futures.Future]) -> None:
        try:
            results = f.result()
            len_results = len(results)
            len_futures = len(futmap)
            if len(results) != len_futures:
                raise RuntimeError(f"Results length {len_results} is different from future-map length {len_futures}")
            for key, value in results.items():
                fut = futmap.get(key, None)
                if fut is None:
                    raise RuntimeError(f"Key {key} not found in future-map: {futmap}")
                if isinstance(value, KafkaError):
                    fut.set_exception(KafkaException(value))
                else:
                    fut.set_result(value)
        except Exception as e:
            for _, fut in futmap.items():
                fut.set_exception(e)

    @staticmethod
    def _create_future() -> concurrent.futures.Future:
        f: concurrent.futures.Future = concurrent.futures.Future()
        if not f.set_running_or_notify_cancel():
            raise RuntimeError("Future was cancelled prematurely")
        return f

    @staticmethod
    def _make_futures(
        futmap_keys: List[Any], class_check: Optional[type], make_result_fn: Any
    ) -> Tuple[concurrent.futures.Future, Dict[Any, concurrent.futures.Future]]:
        """
        Create futures and a futuremap for the keys in futmap_keys,
        and create a request-level future to be bassed to the C API.

        FIXME: use _make_futures_v2 with TypeError in next major release.
        """
        futmap = {}
        for key in futmap_keys:
            if class_check is not None and not isinstance(key, class_check):
                raise ValueError("Expected list of {}".format(repr(class_check)))
            futmap[key] = AdminClient._create_future()

        # Create an internal future for the entire request,
        # this future will trigger _make_..._result() and set result/exception
        # per topic,future in futmap.
        f = AdminClient._create_future()
        f.add_done_callback(lambda f: make_result_fn(f, futmap))

        return f, futmap

    @staticmethod
    def _make_futures_v2(
        futmap_keys: Union[List[Any], Set[Any]], class_check: Optional[type], make_result_fn: Any
    ) -> Tuple[concurrent.futures.Future, Dict[Any, concurrent.futures.Future]]:
        """
        Create futures and a futuremap for the keys in futmap_keys,
        and create a request-level future to be bassed to the C API.
        """
        futmap = {}
        for key in futmap_keys:
            if class_check is not None and not isinstance(key, class_check):
                raise TypeError("Expected list of {}".format(repr(class_check)))
            futmap[key] = AdminClient._create_future()

        # Create an internal future for the entire request,
        # this future will trigger _make_..._result() and set result/exception
        # per topic,future in futmap.
        f = AdminClient._create_future()
        f.add_done_callback(lambda f: make_result_fn(f, futmap))

        return f, futmap

    @staticmethod
    def _make_single_future_pair() -> Tuple[concurrent.futures.Future, concurrent.futures.Future]:
        """
        Create an pair of futures, one for internal usage and one
        to use externally, the external one throws a KafkaException if
        any of the values in the map returned by the first future is
        a KafkaError.
        """

        def single_future_result(internal_f: concurrent.futures.Future, f: concurrent.futures.Future) -> None:
            try:
                results = internal_f.result()
                for _, value in results.items():
                    if isinstance(value, KafkaError):
                        f.set_exception(KafkaException(value))
                        return
                f.set_result(results)
            except Exception as e:
                f.set_exception(e)

        f = AdminClient._create_future()
        internal_f = AdminClient._create_future()
        internal_f.add_done_callback(lambda internal_f: single_future_result(internal_f, f))
        return internal_f, f

    @staticmethod
    def _has_duplicates(items: List[Any]) -> bool:
        return len(set(items)) != len(items)

    @staticmethod
    def _check_list_consumer_group_offsets_request(request: Optional[List[_ConsumerGroupTopicPartitions]]) -> None:
        if request is None:
            raise TypeError("request cannot be None")
        if not isinstance(request, list):
            raise TypeError("request must be a list")
        if len(request) != 1:
            raise ValueError("Currently we support listing offsets for a single consumer group only")
        for req in request:
            if not isinstance(req, _ConsumerGroupTopicPartitions):
                raise TypeError("Expected list of 'ConsumerGroupTopicPartitions'")

            if req.group_id is None:
                raise TypeError("'group_id' cannot be None")
            if not isinstance(req.group_id, string_type):
                raise TypeError("'group_id' must be a string")
            if not req.group_id:
                raise ValueError("'group_id' cannot be empty")

            if req.topic_partitions is not None:
                if not isinstance(req.topic_partitions, list):
                    raise TypeError("'topic_partitions' must be a list or None")
                if len(req.topic_partitions) == 0:
                    raise ValueError("'topic_partitions' cannot be empty")
                for topic_partition in req.topic_partitions:
                    if topic_partition is None:
                        raise ValueError("Element of 'topic_partitions' cannot be None")
                    if not isinstance(topic_partition, _TopicPartition):
                        raise TypeError("Element of 'topic_partitions' must be of type TopicPartition")
                    if topic_partition.topic is None:
                        raise TypeError("Element of 'topic_partitions' must not have 'topic' attribute as None")
                    if not topic_partition.topic:
                        raise ValueError("Element of 'topic_partitions' must not have 'topic' attribute as Empty")
                    if topic_partition.partition < 0:
                        raise ValueError("Element of 'topic_partitions' must not have negative 'partition' value")
                    if topic_partition.offset != OFFSET_INVALID:
                        raise ValueError("Element of 'topic_partitions' must not have 'offset' value")

    @staticmethod
    def _check_alter_consumer_group_offsets_request(request: Optional[List[_ConsumerGroupTopicPartitions]]) -> None:
        if request is None:
            raise TypeError("request cannot be None")
        if not isinstance(request, list):
            raise TypeError("request must be a list")
        if len(request) != 1:
            raise ValueError("Currently we support altering offsets for a single consumer group only")
        for req in request:
            if not isinstance(req, _ConsumerGroupTopicPartitions):
                raise TypeError("Expected list of 'ConsumerGroupTopicPartitions'")
            if req.group_id is None:
                raise TypeError("'group_id' cannot be None")
            if not isinstance(req.group_id, string_type):
                raise TypeError("'group_id' must be a string")
            if not req.group_id:
                raise ValueError("'group_id' cannot be empty")
            if req.topic_partitions is None:
                raise ValueError("'topic_partitions' cannot be null")
            if not isinstance(req.topic_partitions, list):
                raise TypeError("'topic_partitions' must be a list")
            if len(req.topic_partitions) == 0:
                raise ValueError("'topic_partitions' cannot be empty")
            for topic_partition in req.topic_partitions:
                if topic_partition is None:
                    raise ValueError("Element of 'topic_partitions' cannot be None")
                if not isinstance(topic_partition, _TopicPartition):
                    raise TypeError("Element of 'topic_partitions' must be of type TopicPartition")
                if topic_partition.topic is None:
                    raise TypeError("Element of 'topic_partitions' must not have 'topic' attribute as None")
                if not topic_partition.topic:
                    raise ValueError("Element of 'topic_partitions' must not have 'topic' attribute as Empty")
                if topic_partition.partition < 0:
                    raise ValueError("Element of 'topic_partitions' must not have negative value for 'partition' field")
                if topic_partition.offset < 0:
                    raise ValueError("Element of 'topic_partitions' must not have negative value for 'offset' field")

    @staticmethod
    def _check_describe_user_scram_credentials_request(users: Optional[List[str]]) -> None:
        if users is None:
            return
        if not isinstance(users, list):
            raise TypeError("Expected input to be list of String")
        for user in users:
            if user is None:
                raise TypeError("'user' cannot be None")
            if not isinstance(user, string_type):
                raise TypeError("Each value should be a string")
            if not user:
                raise ValueError("'user' cannot be empty")

    @staticmethod
    def _check_alter_user_scram_credentials_request(alterations: List[UserScramCredentialAlteration]) -> None:
        if not isinstance(alterations, list):
            raise TypeError("Expected input to be list")
        if len(alterations) == 0:
            raise ValueError("Expected at least one alteration")
        for alteration in alterations:
            if not isinstance(alteration, UserScramCredentialAlteration):
                raise TypeError("Expected each element of list to be subclass of UserScramCredentialAlteration")
            if alteration.user is None:
                raise TypeError("'user' cannot be None")
            if not isinstance(alteration.user, string_type):
                raise TypeError("'user' must be a string")
            if not alteration.user:
                raise ValueError("'user' cannot be empty")

            if isinstance(alteration, UserScramCredentialUpsertion):
                if alteration.password is None:
                    raise TypeError("'password' cannot be None")
                if not isinstance(alteration.password, bytes):
                    raise TypeError("'password' must be bytes")
                if not alteration.password:
                    raise ValueError("'password' cannot be empty")

                if alteration.salt is not None and not alteration.salt:
                    raise ValueError("'salt' can be None but cannot be empty")
                if alteration.salt and not isinstance(alteration.salt, bytes):
                    raise TypeError("'salt' must be bytes")

                if not isinstance(alteration.scram_credential_info, ScramCredentialInfo):
                    raise TypeError("Expected credential_info to be ScramCredentialInfo Type")
                if alteration.scram_credential_info.iterations < 1:
                    raise ValueError("Iterations should be positive")
                if not isinstance(alteration.scram_credential_info.mechanism, ScramMechanism):
                    raise TypeError("Expected the mechanism to be ScramMechanism Type")
            elif isinstance(alteration, UserScramCredentialDeletion):
                if not isinstance(alteration.mechanism, ScramMechanism):
                    raise TypeError("Expected the mechanism to be ScramMechanism Type")
            else:
                raise TypeError(
                    "Expected each element of list 'alterations' "
                    + "to be either a UserScramCredentialUpsertion or a "
                    + "UserScramCredentialDeletion"
                )

    @staticmethod
    def _check_list_offsets_request(
        topic_partition_offsets: Dict[_TopicPartition, OffsetSpec], kwargs: Dict[str, Any]
    ) -> None:
        if not isinstance(topic_partition_offsets, dict):
            raise TypeError(
                "Expected topic_partition_offsets to be "
                + "dict of [TopicPartitions,OffsetSpec] for list offsets request"
            )

        for topic_partition, offset_spec in topic_partition_offsets.items():
            if topic_partition is None:
                raise TypeError("partition cannot be None")
            if not isinstance(topic_partition, _TopicPartition):
                raise TypeError("partition must be a TopicPartition")
            if topic_partition.topic is None:
                raise TypeError("partition topic name cannot be None")
            if not isinstance(topic_partition.topic, string_type):
                raise TypeError("partition topic name must be string")
            if not topic_partition.topic:
                raise ValueError("partition topic name cannot be empty")
            if topic_partition.partition < 0:
                raise ValueError("partition index must be non-negative")
            if offset_spec is None:
                raise TypeError("OffsetSpec cannot be None")
            if not isinstance(offset_spec, OffsetSpec):
                raise TypeError("Value must be a OffsetSpec")

        if 'isolation_level' in kwargs:
            if not isinstance(kwargs['isolation_level'], _IsolationLevel):
                raise TypeError("isolation_level argument should be an IsolationLevel")

    @staticmethod
    def _check_delete_records(request: List[_TopicPartition]) -> None:
        if not isinstance(request, list):
            raise TypeError(f"Expected Request to be a list, got '{type(request).__name__}' ")
        for req in request:
            if not isinstance(req, _TopicPartition):
                raise TypeError(
                    "Element of the request list must be of type 'TopicPartition'" + f" got '{type(req).__name__}' "
                )
            if req.partition < 0:
                raise ValueError("'partition' cannot be negative")

    @staticmethod
    def _check_elect_leaders(election_type: _ElectionType, partitions: Optional[List[_TopicPartition]]) -> None:
        if not isinstance(election_type, _ElectionType):
            raise TypeError("Expected 'election_type' to be of type 'ElectionType'")
        if partitions is not None:
            if not isinstance(partitions, list):
                raise TypeError("Expected 'partitions' to be a list, got " + f"'{type(partitions).__name__}'")
            for partition in partitions:
                if not isinstance(partition, _TopicPartition):
                    raise TypeError(
                        "Element of the 'partitions' list must be of type 'TopicPartition'"
                        + f" got '{type(partition).__name__}' "
                    )
                if partition.partition < 0:
                    raise ValueError(
                        "Elements of the 'partitions' list must not have negative value" + " for 'partition' field"
                    )

    def create_topics(  # type: ignore[override]
        self, new_topics: List[NewTopic], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Create one or more new topics.

        :param list(NewTopic) new_topics: A list of specifictions (NewTopic) for
                  the topics that should be created.
        :param float operation_timeout: The operation timeout in seconds,
                  controlling how long the CreateTopics request will block
                  on the broker waiting for the topic creation to propagate
                  in the cluster. A value of 0 returns immediately.
                  Default: `socket.timeout.ms/1000.0`
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`
        :param bool validate_only: If true, the request is only validated
                  without creating the topic. Default: False

        :returns: A dict of futures for each topic, keyed by the topic name.
                  The future result() method returns None.

        :rtype: dict(<topic_name, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        f, futmap = AdminClient._make_futures([x.topic for x in new_topics], None, AdminClient._make_topics_result)

        super(AdminClient, self).create_topics(new_topics, f, **kwargs)

        return futmap

    def delete_topics(  # type: ignore[override]
        self, topics: List[str], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Delete one or more topics.

        :param list(str) topics: A list of topics to mark for deletion.
        :param float operation_timeout: The operation timeout in seconds,
                  controlling how long the DeleteTopics request will block
                  on the broker waiting for the topic deletion to propagate
                  in the cluster. A value of 0 returns immediately.
                  Default: `socket.timeout.ms/1000.0`
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each topic, keyed by the topic name.
                  The future result() method returns None.

        :rtype: dict(<topic_name, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        f, futmap = AdminClient._make_futures(topics, None, AdminClient._make_topics_result)

        super(AdminClient, self).delete_topics(topics, f, **kwargs)

        return futmap

    def list_topics(self, *args: Any, **kwargs: Any) -> ClusterMetadata:
        return super(AdminClient, self).list_topics(*args, **kwargs)

    def list_groups(self, *args: Any, **kwargs: Any) -> List[GroupMetadata]:
        return super(AdminClient, self).list_groups(*args, **kwargs)

    def create_partitions(  # type: ignore[override]
        self, new_partitions: List[NewPartitions], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Create additional partitions for the given topics.

        :param list(NewPartitions) new_partitions: New partitions to be created.
        :param float operation_timeout: The operation timeout in seconds,
                  controlling how long the CreatePartitions request will block
                  on the broker waiting for the partition creation to propagate
                  in the cluster. A value of 0 returns immediately.
                  Default: `socket.timeout.ms/1000.0`
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`
        :param bool validate_only: If true, the request is only validated
                  without creating the partitions. Default: False

        :returns: A dict of futures for each topic, keyed by the topic name.
                  The future result() method returns None.

        :rtype: dict(<topic_name, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        f, futmap = AdminClient._make_futures([x.topic for x in new_partitions], None, AdminClient._make_topics_result)

        super(AdminClient, self).create_partitions(new_partitions, f, **kwargs)

        return futmap

    def describe_configs(  # type: ignore[override]
        self, resources: List[ConfigResource], **kwargs: Any
    ) -> Dict[ConfigResource, concurrent.futures.Future]:
        """
        Get the configuration of the specified resources.

        :warning: Multiple resources and resource types may be requested,
                  but at most one resource of type RESOURCE_BROKER is allowed
                  per call since these resource requests must be sent to the
                  broker specified in the resource.

        :param list(ConfigResource) resources: Resources to get the configuration for.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each resource, keyed by the ConfigResource.
                  The type of the value returned by the future result() method is
                  dict(<configname, ConfigEntry>).

        :rtype: dict(<ConfigResource, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        f, futmap = AdminClient._make_futures(resources, ConfigResource, AdminClient._make_resource_result)

        super(AdminClient, self).describe_configs(resources, f, **kwargs)

        return futmap

    def alter_configs(  # type: ignore[override]
        self, resources: List[ConfigResource], **kwargs: Any
    ) -> Dict[ConfigResource, concurrent.futures.Future]:
        """
        .. deprecated:: 2.2.0

        Update configuration properties for the specified resources.
        Updates are not transactional so they may succeed for a subset
        of the provided resources while the others fail.
        The configuration for a particular resource is updated atomically,
        replacing the specified values while reverting unspecified configuration
        entries to their default values.

        :warning: alter_configs() will replace all existing configuration for
                  the provided resources with the new configuration given,
                  reverting all other configuration for the resource back
                  to their default values.

        :warning: Multiple resources and resource types may be specified,
                  but at most one resource of type RESOURCE_BROKER is allowed
                  per call since these resource requests must be sent to the
                  broker specified in the resource.

        :param list(ConfigResource) resources: Resources to update configuration of.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`.
        :param bool validate_only: If true, the request is validated only,
                  without altering the configuration. Default: False

        :returns: A dict of futures for each resource, keyed by the ConfigResource.
                  The future result() method returns None or throws a KafkaException.

        :rtype: dict(<ConfigResource, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid type.
        :raises ValueError: Invalid value.
        """
        warnings.warn(
            "alter_configs has been deprecated. Use incremental_alter_configs instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        f, futmap = AdminClient._make_futures(resources, ConfigResource, AdminClient._make_resource_result)

        super(AdminClient, self).alter_configs(resources, f, **kwargs)

        return futmap

    def incremental_alter_configs(  # type: ignore[override]
        self, resources: List[ConfigResource], **kwargs: Any
    ) -> Dict[ConfigResource, concurrent.futures.Future]:
        """
        Update configuration properties for the specified resources.
        Updates are incremental, i.e only the values mentioned are changed
        and rest remain as is.

        :param list(ConfigResource) resources: Resources to update configuration of.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`.
        :param bool validate_only: If true, the request is validated only,
                  without altering the configuration. Default: False
        :param int broker: Broker id to send the request to. When
                  altering broker configurations, it's ignored because
                  the request needs to go to that broker only.
                  Default: controller broker.

        :returns: A dict of futures for each resource, keyed by the ConfigResource.
                  The future result() method returns None or throws a KafkaException.

        :rtype: dict(<ConfigResource, future>)

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid type.
        :raises ValueError: Invalid value.
        """
        f, futmap = AdminClient._make_futures_v2(resources, ConfigResource, AdminClient._make_resource_result)

        super(AdminClient, self).incremental_alter_configs(resources, f, **kwargs)

        return futmap

    def create_acls(  # type: ignore[override]
        self, acls: List[AclBinding], **kwargs: Any
    ) -> Dict[AclBinding, concurrent.futures.Future]:
        """
        Create one or more ACL bindings.

        :param list(AclBinding) acls: A list of unique ACL binding specifications (:class:`.AclBinding`)
                         to create.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each ACL binding, keyed by the :class:`AclBinding` object.
                  The future result() method returns None on success.

        :rtype: dict[AclBinding, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """
        if AdminClient._has_duplicates(acls):
            raise ValueError("duplicate ACL bindings not allowed")

        f, futmap = AdminClient._make_futures(acls, AclBinding, AdminClient._make_acls_result)

        super(AdminClient, self).create_acls(acls, f, **kwargs)

        return futmap

    def describe_acls(  # type: ignore[override]
        self, acl_binding_filter: AclBindingFilter, **kwargs: Any
    ) -> concurrent.futures.Future:
        """
        Match ACL bindings by filter.

        :param AclBindingFilter acl_binding_filter: a filter with attributes that
                  must match.
                  String attributes match exact values or any string if set to None.
                  Enums attributes match exact values or any value if equal to `ANY`.
                  If :class:`ResourcePatternType` is set to :attr:`ResourcePatternType.MATCH`
                  returns ACL bindings with:
                  :attr:`ResourcePatternType.LITERAL` pattern type with resource name equal
                  to the given resource name;
                  :attr:`ResourcePatternType.LITERAL` pattern type with wildcard resource name
                  that matches the given resource name;
                  :attr:`ResourcePatternType.PREFIXED` pattern type with resource name
                  that is a prefix of the given resource name
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A future returning a list(:class:`AclBinding`) as result

        :rtype: future

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        f = AdminClient._create_future()

        super(AdminClient, self).describe_acls(acl_binding_filter, f, **kwargs)

        return f

    def delete_acls(  # type: ignore[override]
        self, acl_binding_filters: List[AclBindingFilter], **kwargs: Any
    ) -> Dict[AclBindingFilter, concurrent.futures.Future]:
        """
        Delete ACL bindings matching one or more ACL binding filters.

        :param list(AclBindingFilter) acl_binding_filters: a list of unique ACL binding filters
                  to match ACLs to delete.
                  String attributes match exact values or any string if set to None.
                  Enums attributes match exact values or any value if equal to `ANY`.
                  If :class:`ResourcePatternType` is set to :attr:`ResourcePatternType.MATCH`
                  deletes ACL bindings with:
                  :attr:`ResourcePatternType.LITERAL` pattern type with resource name
                  equal to the given resource name;
                  :attr:`ResourcePatternType.LITERAL` pattern type with wildcard resource name
                  that matches the given resource name;
                  :attr:`ResourcePatternType.PREFIXED` pattern type with resource name
                  that is a prefix of the given resource name
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each ACL binding filter, keyed by the :class:`AclBindingFilter` object.
                  The future result() method returns a list of :class:`AclBinding`.

        :rtype: dict[AclBindingFilter, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """
        if AdminClient._has_duplicates(acl_binding_filters):
            raise ValueError("duplicate ACL binding filters not allowed")

        f, futmap = AdminClient._make_futures(acl_binding_filters, AclBindingFilter, AdminClient._make_acls_result)

        super(AdminClient, self).delete_acls(acl_binding_filters, f, **kwargs)

        return futmap

    def list_consumer_groups(self, **kwargs: Any) -> concurrent.futures.Future:  # type: ignore[override]
        """
        List consumer groups.

        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`
        :param set(ConsumerGroupState) states: only list consumer groups which are currently in
                  these states.
        :param set(ConsumerGroupType) types: only list consumer groups of
                  these types.

        :returns: a future. Result method of the future returns :class:`ListConsumerGroupsResult`.

        :rtype: future

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """
        if "states" in kwargs:
            states = kwargs["states"]
            if states is not None:
                if not isinstance(states, set):
                    raise TypeError("'states' must be a set")
                for state in states:
                    if not isinstance(state, _ConsumerGroupState):
                        raise TypeError("All elements of states must be of type ConsumerGroupState")
                kwargs["states_int"] = [state.value for state in states]
            kwargs.pop("states")
        if "types" in kwargs:
            types = kwargs["types"]
            if types is not None:
                if not isinstance(types, set):
                    raise TypeError("'types' must be a set")
                for type in types:
                    if not isinstance(type, _ConsumerGroupType):
                        raise TypeError("All elements of types must be of type ConsumerGroupType")
                kwargs["types_int"] = [type.value for type in types]
            kwargs.pop("types")

        f, _ = AdminClient._make_futures([], None, AdminClient._make_list_consumer_groups_result)

        super(AdminClient, self).list_consumer_groups(f, **kwargs)

        return f

    def describe_consumer_groups(  # type: ignore[override]
        self, group_ids: List[str], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Describe consumer groups.

        :param list(str) group_ids: List of group_ids which need to be described.
        :param bool include_authorized_operations: If True, fetches group AclOperations. Default: False
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each group, keyed by the group_id.
                  The future result() method returns :class:`ConsumerGroupDescription`.

        :rtype: dict[str, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        if not isinstance(group_ids, list):
            raise TypeError("Expected input to be list of group ids to be described")

        if len(group_ids) == 0:
            raise ValueError("Expected at least one group to be described")

        f, futmap = AdminClient._make_futures(group_ids, None, AdminClient._make_consumer_groups_result)

        super(AdminClient, self).describe_consumer_groups(group_ids, f, **kwargs)

        return futmap

    def describe_topics(  # type: ignore[override]
        self, topics: _TopicCollection, **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Describe topics.

        :param TopicCollection topics: Collection of list of topic names to describe.
        :param bool include_authorized_operations: If True, fetches topic AclOperations. Default: False
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each topic, keyed by the topic.
                  The future result() method returns :class:`TopicDescription`.

        :rtype: dict[str, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """

        if not isinstance(topics, _TopicCollection):
            raise TypeError("Expected input to be instance of TopicCollection")

        topic_names = topics.topic_names

        if not isinstance(topic_names, list):
            raise TypeError("Expected list of topic names to be described")

        f, futmap = AdminClient._make_futures_v2(topic_names, None, AdminClient._make_futmap_result_from_list)

        super(AdminClient, self).describe_topics(topic_names, f, **kwargs)  # type: ignore[arg-type]

        return futmap

    def describe_cluster(self, **kwargs: Any) -> concurrent.futures.Future:  # type: ignore[override]
        """
        Describe cluster.

        :param bool include_authorized_operations: If True, fetches topic AclOperations. Default: False
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A future returning description of the cluster as result

        :rtype: future containing the description of the cluster in result.

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """

        f = AdminClient._create_future()

        super(AdminClient, self).describe_cluster(f, **kwargs)

        return f

    def delete_consumer_groups(  # type: ignore[override]
        self, group_ids: List[str], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Delete the given consumer groups.

        :param list(str) group_ids: List of group_ids which need to be deleted.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each group, keyed by the group_id.
                  The future result() method returns None.

        :rtype: dict[str, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """
        if not isinstance(group_ids, list):
            raise TypeError("Expected input to be list of group ids to be deleted")

        if len(group_ids) == 0:
            raise ValueError("Expected at least one group to be deleted")

        f, futmap = AdminClient._make_futures(group_ids, string_type, AdminClient._make_consumer_groups_result)

        super(AdminClient, self).delete_consumer_groups(group_ids, f, **kwargs)

        return futmap

    def list_consumer_group_offsets(  # type: ignore[override]
        self, list_consumer_group_offsets_request: List[_ConsumerGroupTopicPartitions], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        List offset information for the consumer group and (optional) topic partition provided in the request.

        :note: Currently, the API supports only a single group.

        :param list(ConsumerGroupTopicPartitions) list_consumer_group_offsets_request: List of
                    :class:`ConsumerGroupTopicPartitions` which consist of group name and topic
                    partition information for which offset detail is expected. If only group name is
                    provided, then offset information of all the topic and partition associated with
                    that group is returned.
        :param bool require_stable: If True, fetches stable offsets. Default: False
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each group, keyed by the group id.
                  The future result() method returns :class:`ConsumerGroupTopicPartitions`.

        :rtype: dict[str, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        AdminClient._check_list_consumer_group_offsets_request(list_consumer_group_offsets_request)

        f, futmap = AdminClient._make_futures(
            [request.group_id for request in list_consumer_group_offsets_request],
            string_type,
            AdminClient._make_consumer_group_offsets_result,
        )

        super(AdminClient, self).list_consumer_group_offsets(list_consumer_group_offsets_request, f, **kwargs)

        return futmap

    def alter_consumer_group_offsets(  # type: ignore[override]
        self, alter_consumer_group_offsets_request: List[_ConsumerGroupTopicPartitions], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Alter offset for the consumer group and topic partition provided in the request.

        :note: Currently, the API supports only a single group.

        :param list(ConsumerGroupTopicPartitions) alter_consumer_group_offsets_request: List of
                    :class:`ConsumerGroupTopicPartitions` which consist of group name and topic
                    partition; and corresponding offset to be updated.
        :param float request_timeout: The overall request timeout in seconds,
                  including broker lookup, request transmission, operation time
                  on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures for each group, keyed by the group id.
                  The future result() method returns :class:`ConsumerGroupTopicPartitions`.

        :rtype: dict[ConsumerGroupTopicPartitions, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        :raises ValueException: Invalid input.
        """

        AdminClient._check_alter_consumer_group_offsets_request(alter_consumer_group_offsets_request)

        f, futmap = AdminClient._make_futures(
            [request.group_id for request in alter_consumer_group_offsets_request],
            string_type,
            AdminClient._make_consumer_group_offsets_result,
        )

        super(AdminClient, self).alter_consumer_group_offsets(alter_consumer_group_offsets_request, f, **kwargs)

        return futmap

    def set_sasl_credentials(self, username: str, password: str) -> None:
        """
        Sets the SASL credentials used for this client.
        These credentials will overwrite the old ones, and will be used the
        next time the client needs to authenticate.
        This method will not disconnect existing broker connections that
        have been established with the old credentials.
        This method is applicable only to SASL PLAIN and SCRAM mechanisms.

        :param str username: The username to set.
        :param str password: The password to set.

        :rtype: None

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeException: Invalid input.
        """
        super(AdminClient, self).set_sasl_credentials(username, password)

    def describe_user_scram_credentials(  # type: ignore[override]
        self, users: Optional[List[str]] = None, **kwargs: Any
    ) -> Union[concurrent.futures.Future, Dict[str, concurrent.futures.Future]]:
        """
        Describe user SASL/SCRAM credentials.

        :param list(str) users: List of user names to describe.
               Duplicate users aren't allowed. Can be None
               to describe all user's credentials.
        :param float request_timeout: The overall request timeout in seconds,
               including broker lookup, request transmission, operation time
               on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: In case None is passed it returns a single future.
                  The future yields a dict[str, UserScramCredentialsDescription]
                  or raises a KafkaException

                  In case a list of user names is passed, it returns
                  a dict[str, future[UserScramCredentialsDescription]].
                  The futures yield a :class:`UserScramCredentialsDescription`
                  or raise a KafkaException

        :rtype: Union[future[dict[str, UserScramCredentialsDescription]],
                      dict[str, future[UserScramCredentialsDescription]]]

        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """
        AdminClient._check_describe_user_scram_credentials_request(users)

        if users is None:
            internal_f, ret_fut = AdminClient._make_single_future_pair()
        else:
            internal_f, ret_fut = AdminClient._make_futures_v2(  # type: ignore[assignment]
                users, None, AdminClient._make_futmap_result
            )
        super(AdminClient, self).describe_user_scram_credentials(users, internal_f, **kwargs)
        return ret_fut

    def alter_user_scram_credentials(  # type: ignore[override]
        self, alterations: List[UserScramCredentialAlteration], **kwargs: Any
    ) -> Dict[str, concurrent.futures.Future]:
        """
        Alter user SASL/SCRAM credentials.

        :param list(UserScramCredentialAlteration) alterations: List of
               :class:`UserScramCredentialAlteration` to apply.
               The pair (user, mechanism) must be unique among alterations.
        :param float request_timeout: The overall request timeout in seconds,
               including broker lookup, request transmission, operation time
               on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures keyed by user name.
                  The future result() method returns None or
                  raises KafkaException

        :rtype: dict[str, future]

        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """
        AdminClient._check_alter_user_scram_credentials_request(alterations)

        f, futmap = AdminClient._make_futures_v2(
            set([alteration.user for alteration in alterations]), None, AdminClient._make_futmap_result
        )

        super(AdminClient, self).alter_user_scram_credentials(alterations, f, **kwargs)
        return futmap

    def list_offsets(  # type: ignore[override]
        self, topic_partition_offsets: Dict[_TopicPartition, OffsetSpec], **kwargs: Any
    ) -> Dict[_TopicPartition, concurrent.futures.Future]:
        """
        Enables to find the beginning offset,
        end offset as well as the offset matching a timestamp
        or the offset with max timestamp in partitions.

        :param dict([TopicPartition, OffsetSpec]) topic_partition_offsets: Dictionary of
               TopicPartition objects associated with the corresponding OffsetSpec to query for.
        :param IsolationLevel isolation_level: The isolation level to use when
               querying.
        :param float request_timeout: The overall request timeout in seconds,
               including broker lookup, request transmission, operation time
               on broker, and response. Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures keyed by TopicPartition.
                  The future result() method returns ListOffsetsResultInfo
                  raises KafkaException

        :rtype: dict[TopicPartition, future]

        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """
        AdminClient._check_list_offsets_request(topic_partition_offsets, kwargs)

        if 'isolation_level' in kwargs:
            kwargs['isolation_level_value'] = kwargs['isolation_level'].value
            del kwargs['isolation_level']

        topic_partition_offsets_list = [
            _TopicPartition(topic_partition.topic, int(topic_partition.partition), int(offset_spec._value))
            for topic_partition, offset_spec in topic_partition_offsets.items()
        ]

        f, futmap = AdminClient._make_futures_v2(
            topic_partition_offsets_list, _TopicPartition, AdminClient._make_futmap_result
        )

        super(AdminClient, self).list_offsets(topic_partition_offsets_list, f, **kwargs)
        return futmap

    def delete_records(  # type: ignore[override]
        self, topic_partition_offsets: List[_TopicPartition], **kwargs: Any
    ) -> Dict[_TopicPartition, concurrent.futures.Future]:
        """
        Deletes all the records before the specified offsets (not including),
        in the specified topics and partitions.

        :param list(TopicPartition) topic_partition_offsets: A list of
               :class:`.TopicPartition` objects having `offset` field set to the offset
               before which all the records should be deleted.
               `offset` can be set to :py:const:`OFFSET_END` (-1) to delete all records
               in the partition.
        :param float request_timeout: The overall request timeout in seconds,
               including broker lookup, request transmission, operation time
               on broker, and response. Default: `socket.timeout.ms/1000.0`
        :param float operation_timeout: The operation timeout in seconds,
               controlling how long the `delete_records` request will block
               on the broker waiting for the record deletion to propagate
               in the cluster. A value of 0 returns immediately.
               Default: `socket.timeout.ms/1000.0`

        :returns: A dict of futures keyed by the :class:`.TopicPartition`.
                  The future result() method returns :class:`.DeletedRecords`
                  or raises :class:`.KafkaException`

        :rtype: dict[TopicPartition, future]

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """
        AdminClient._check_delete_records(topic_partition_offsets)

        f, futmap = AdminClient._make_futures_v2(
            topic_partition_offsets, _TopicPartition, AdminClient._make_futmap_result
        )

        super(AdminClient, self).delete_records(topic_partition_offsets, f, **kwargs)
        return futmap

    def elect_leaders(  # type: ignore[override]
        self, election_type: _ElectionType, partitions: Optional[List[_TopicPartition]] = None, **kwargs: Any
    ) -> concurrent.futures.Future:
        """
        Perform Preferred or Unclean leader election for
        all the specified partitions or all partitions in the cluster.

        :param ElectionType election_type: The type of election to perform.
        :param List[TopicPartition]|None partitions: The topic partitions to perform
               the election on. Use ``None`` to perform on all the topic partitions.
        :param float request_timeout: The overall request timeout in seconds,
               including broker lookup, request transmission, operation time
               on broker, and response. Default: `socket.timeout.ms*1000.0`
        :param float operation_timeout: The operation timeout in seconds,
               controlling how long the 'elect_leaders' request will block
               on the broker waiting for the election to propagate
               in the cluster. A value of 0 returns immediately.
               Default: `socket.timeout.ms/1000.0`

        :returns: A future. Method result() of the future returns
                  dict[TopicPartition, KafkaException|None].

        :rtype: future

        :raises KafkaException: Operation failed locally or on broker.
        :raises TypeError: Invalid input type.
        :raises ValueError: Invalid input value.
        """

        AdminClient._check_elect_leaders(election_type, partitions)

        f = AdminClient._create_future()

        super(AdminClient, self).elect_leaders(election_type.value, partitions, f, **kwargs)

        return f
